"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

import base64
from typing import Any, List

from botbuilder.core import TurnContext

from ...input_file import InputFile
from ...state import MemoryBase
from ..tokenizers import Tokenizer
from .message import (
    ImageContentPart,
    ImageUrl,
    Message,
    MessageContentParts,
    TextContentPart,
)
from .prompt_functions import PromptFunctions
from .rendered_prompt_section import RenderedPromptSection
from .sections.prompt_section_base import PromptSectionBase


class UserInputMessage(PromptSectionBase):
    """
    A section capable of rendering user input text and images as a user message.
    """

    _input_variable: str
    _files_variable: str

    def __init__(
        self, tokens: float = -1, input_variable: str = "input", files_variable: str = "input_files"
    ):
        """
        Creates a new 'UserInputMessage' instance.

        Args:
            tokens (float, optional): Sizing strategy for this section. Defaults to `auto`.
            input_variable (str, optional): Name of the variable containing the user input text.
              Defaults to `input`.
            files_variable (str, optional): Name of the variable containing the user input files.
              Defaults to `inputFiles`.
        """
        super().__init__(tokens, True, "\n", "user: ")
        self._input_variable = input_variable
        self._files_variable = files_variable

    async def render_as_messages(
        self,
        context: TurnContext,
        memory: MemoryBase,
        functions: PromptFunctions,
        tokenizer: Tokenizer,
        max_tokens: int,
    ) -> RenderedPromptSection[List[Message[Any]]]:
        """
        Renders the section as a list of messages.

        Args:
            context (TurnContext): Context for the current turn of conversation with the user.
            memory (MemoryBase): An interface for accessing state values.
            functions (PromptFunctions): Registry of functions that can be used by the section.
            tokenizer (Tokenizer): Tokenizer to use when rendering the section.
            max_tokens (int): Maximum number of tokens allowed to be rendered.

        Returns:
            RenderedPromptSection[List[Message]]: The rendered prompt section as a list of messages.
        """
        input_text: str = memory.get(self._input_variable) or ""
        input_files: List[InputFile] = memory.get(self._files_variable) or []

        message: Message[List[MessageContentParts]] = Message("user", [])

        length = 0
        budget = self._get_token_budget(max_tokens)
        if len(input_text) > 0:
            encoded = tokenizer.encode(input_text)
            if len(encoded) <= budget:
                if message.content is not None:
                    message.content.append(TextContentPart(type="text", text=input_text))
                    length += len(encoded)
                    budget -= len(encoded)
            else:
                if message.content is not None:
                    message.content.append(
                        TextContentPart(type="text", text=tokenizer.decode(encoded[:budget]))
                    )
                    length += budget
                    budget = 0

        # Append image content parts
        images = [f for f in input_files if f.content_type.startswith("image/")]
        for image in images:
            # Check for budget to add image
            if budget < 85:
                break

            url = (
                f"data:{image.content_type};base64,"
                f"{base64.b64encode(image.content).decode('utf-8')}"
            )
            if message.content is not None:
                message.content.append(ImageContentPart(type="image_url", image_url=ImageUrl(url)))
                length += 85
                budget -= 85

        return RenderedPromptSection(output=[message], length=length, too_long=False)



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from .activity_type import ActivityType, ConversationUpdateType
from .app import Application
from .app_error import ApplicationError
from .app_options import ApplicationOptions
from .input_file import InputFile
from .message_reaction_types import MessageReactionTypes
from .query import Query
from .teams_adapter import TeamsAdapter

__all__ = [
    "ActivityType",
    "ConversationUpdateType",
    "Application",
    "ApplicationError",
    "ApplicationOptions",
    "InputFile",
    "Query",
    "TeamsAdapter",
    "MessageReactionTypes",
]



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from dataclasses import dataclass


@dataclass
class AdaptiveCardsSearchParams:
    query_text: str
    dataset: str



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from typing import List, Literal, Union
from unittest import IsolatedAsyncioTestCase, mock

import httpx
import openai

from teams.ai.embeddings.azure_openai_embeddings import AzureOpenAIEmbeddings
from teams.ai.embeddings.azure_openai_embeddings_options import (
    AzureOpenAIEmbeddingsOptions,
)

embedding_1 = [
    -0.006929283495992422,
    -0.016709936782717705,
    0.009889421984553337,
    -0.0007253705407492816,
]
embedding_2 = [
    -0.006929283495992422,
    -0.016709936782717705,
    0.009889421984553337,
    -0.0007253705407492816,
]


class MockAsyncEmbeddings:
    async def create(
        self,
        input: Union[str, List[str], List[int], List[List[int]]],
        model: Union[str, Literal["text-embedding-ada-002"]],
    ) -> openai.types.CreateEmbeddingResponse:
        # pylint: disable=unused-argument
        # this is necessary to override and mock the class
        return openai.types.CreateEmbeddingResponse(
            data=[
                openai.types.Embedding(embedding=embedding_1, index=1, object="embedding"),
                openai.types.Embedding(embedding=embedding_2, index=0, object="embedding"),
            ],
            model="text-embedding-ada-002",
            object="list",
            usage=openai.types.create_embedding_response.Usage(prompt_tokens=5, total_tokens=5),
        )


class MockAsyncAzureOpenAI:
    embeddings = MockAsyncEmbeddings()


class MockAsyncEmbeddingsRateLimited:
    async def create(
        self,
        input: Union[str, List[str], List[int], List[List[int]]],
        model: Union[str, Literal["text-embedding-ada-002"]],
    ) -> openai.types.CreateEmbeddingResponse:
        raise openai.RateLimitError(
            message="This is a rate limited error",
            response=httpx.Response(
                status_code=429, request=httpx.Request(method="method", url="url")
            ),
            body=None,
        )


class MockAsyncAzureOpenAIRateLimited:
    embeddings = MockAsyncEmbeddingsRateLimited()


class MockAsyncEmbeddingsAPIError:
    async def create(
        self,
        input: Union[str, List[str], List[int], List[List[int]]],
        model: Union[str, Literal["text-embedding-ada-002"]],
    ) -> openai.types.CreateEmbeddingResponse:
        raise openai.APIError(
            message="This is a bad request error",
            request=httpx.Request(method="method", url="url"),
            body=None,
        )


class MockAsyncAzureOpenAIAPIError:
    embeddings = MockAsyncEmbeddingsAPIError()


class TestOpenAIEmbeddings(IsolatedAsyncioTestCase):
    options: AzureOpenAIEmbeddingsOptions
    embeddings: AzureOpenAIEmbeddings
    options_with_array: AzureOpenAIEmbeddingsOptions
    embeddings_with_array: AzureOpenAIEmbeddings

    def setUp(self):
        self.options = AzureOpenAIEmbeddingsOptions(
            azure_api_key="empty",
            azure_endpoint="https://mock.openai.azure.com/",
            azure_deployment="text-embedding-ada-002",
            log_requests=True,
            retry_policy=[2, 4],
        )
        self.options_with_array = AzureOpenAIEmbeddingsOptions(
            azure_api_key="empty",
            azure_endpoint="https://mock.openai.azure.com",
            azure_deployment="text-embedding-ada-002",
            azure_api_version="2023-05-15",
            log_requests=False,
            request_config={
                "Content-Type": "application/json",
                "User-Agent": "123",
            },
        )

    @mock.patch("openai.AsyncAzureOpenAI", return_value=MockAsyncAzureOpenAI)
    async def test_string_embedding(self, mock_async_azure_open_ai):
        self.embeddings = AzureOpenAIEmbeddings(self.options)
        section = await self.embeddings.create_embeddings("This is an embedding")
        self.assertTrue(mock_async_azure_open_ai.called)
        self.assertEqual(section.status, "success")
        self.assertEqual(section.output, [embedding_2, embedding_1])

    @mock.patch("openai.AsyncAzureOpenAI", return_value=MockAsyncAzureOpenAI)
    async def test_array_embedding(self, mock_async_azure_open_ai):
        self.embeddings_with_array = AzureOpenAIEmbeddings(self.options_with_array)
        section = await self.embeddings_with_array.create_embeddings(["This is", "an embedding"])
        self.assertTrue(mock_async_azure_open_ai.called)
        self.assertEqual(section.status, "success")
        self.assertEqual(section.output, [embedding_2, embedding_1])

    @mock.patch("openai.AsyncAzureOpenAI", return_value=MockAsyncAzureOpenAIAPIError)
    async def test_api_error(self, mock_async_azure_open_ai):
        self.embeddings = AzureOpenAIEmbeddings(self.options)
        section = await self.embeddings.create_embeddings("This is an embedding")
        self.assertTrue(mock_async_azure_open_ai.called)
        self.assertEqual(section.status, "error")
        self.assertEqual(
            section.output,
            "The embeddings API returned an error status of None: This is a bad request error",
        )

    @mock.patch("openai.AsyncAzureOpenAI", return_value=MockAsyncAzureOpenAIRateLimited)
    async def test_rate_limited(self, mock_async_azure_open_ai):
        self.embeddings = AzureOpenAIEmbeddings(self.options)
        section = await self.embeddings.create_embeddings("This is an embedding")
        self.assertTrue(mock_async_azure_open_ai.called)
        self.assertEqual(section.status, "rate_limited")
        self.assertEqual(section.output, "The embeddings API returned a rate limit error.")



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from importlib.metadata import version


class _UserAgent:
    @property
    def user_agent(self) -> str:
        return f"teamsai-py/{version('teams-ai')}"



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from typing import cast
from unittest import IsolatedAsyncioTestCase

from botbuilder.core import TurnContext

from teams.ai.models import PromptResponse
from teams.ai.prompts import Message
from teams.ai.tokenizers import GPTTokenizer
from teams.ai.validators import JSONResponseValidator
from teams.state import TurnState


class TestJSONResponseValidator(IsolatedAsyncioTestCase):
    async def test_should_be_invalid_when_no_message(self):
        validator = JSONResponseValidator()
        res = await validator.validate_response(
            context=cast(TurnContext, {}),
            memory=TurnState(),
            tokenizer=GPTTokenizer(),
            response=PromptResponse(),
            remaining_attempts=3,
        )

        self.assertFalse(res.valid)
        self.assertEqual(res.feedback, None)

    async def test_should_be_invalid_when_missing_json(self):
        validator = JSONResponseValidator()
        response = PromptResponse(message=Message(role="assistant", content=""))

        res = await validator.validate_response(
            context=cast(TurnContext, {}),
            memory=TurnState(),
            tokenizer=GPTTokenizer(),
            response=response,
            remaining_attempts=3,
        )

        self.assertFalse(res.valid)
        self.assertEqual(res.feedback, validator.missing_json_feedback)

    async def test_should_be_invalid_when_invalid_json(self):
        validator = JSONResponseValidator()
        response = PromptResponse(message=Message(role="assistant", content="{]"))

        res = await validator.validate_response(
            context=cast(TurnContext, {}),
            memory=TurnState(),
            tokenizer=GPTTokenizer(),
            response=response,
            remaining_attempts=3,
        )

        self.assertFalse(res.valid)
        self.assertEqual(res.feedback, validator.missing_json_feedback)

    async def test_should_be_valid_when_object_in_text(self):
        validator = JSONResponseValidator()
        response = PromptResponse(
            message=Message(
                role="assistant", content='hi this is an object { "hello": "world" } in json!'
            )
        )

        res = await validator.validate_response(
            context=cast(TurnContext, {}),
            memory=TurnState(),
            tokenizer=GPTTokenizer(),
            response=response,
            remaining_attempts=3,
        )

        self.assertTrue(res.valid)
        self.assertEqual(res.value, {"hello": "world"})

    async def test_should_be_valid_when_multiple_object_in_text_and_return_last_object(self):
        validator = JSONResponseValidator()
        response = PromptResponse(
            message=Message(
                role="assistant",
                content="""
            The following are some objects:

            { \"a\": \"b\" }
            { \"c\": \"d\" }
            [{ \"e\": \"f\" }]

            enjoy!
            """,
            )
        )

        res = await validator.validate_response(
            context=cast(TurnContext, {}),
            memory=TurnState(),
            tokenizer=GPTTokenizer(),
            response=response,
            remaining_attempts=3,
        )

        self.assertTrue(res.valid)
        self.assertEqual(res.value, {"e": "f"})

    async def test_should_be_invalid_when_object_doesnt_match_schema(self):
        validator = JSONResponseValidator(
            schema={
                "type": "object",
                "properties": {"name": {"type": "string"}, "age": {"type": "number"}},
                "required": ["name", "age"],
            }
        )

        response = PromptResponse(
            message=Message(
                role="assistant", content='hi this is an object { "hello": "world" } in json!'
            )
        )

        res = await validator.validate_response(
            context=cast(TurnContext, {}),
            memory=TurnState(),
            tokenizer=GPTTokenizer(),
            response=response,
            remaining_attempts=3,
        )

        self.assertFalse(res.valid)
        self.assertEqual(res.feedback, f"{validator.error_feedback}\n'name' is a required property")

    async def test_should_be_valid_when_object_matchs_schema(self):
        validator = JSONResponseValidator(
            schema={
                "type": "object",
                "properties": {"name": {"type": "string"}, "age": {"type": "number"}},
                "required": ["name", "age"],
            }
        )

        response = PromptResponse(
            message=Message(
                role="assistant",
                content='hi this is an object { "name": "test", "age": 10 } in json!',
            )
        )

        res = await validator.validate_response(
            context=cast(TurnContext, {}),
            memory=TurnState(),
            tokenizer=GPTTokenizer(),
            response=response,
            remaining_attempts=3,
        )

        self.assertTrue(res.valid)
        self.assertEqual(res.feedback, None)
        self.assertEqual(res.value, {"name": "test", "age": 10})



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from typing import List

from botbuilder.core import TurnContext

from ....state import MemoryBase
from ...data_sources import DataSource
from ...tokenizers import Tokenizer
from ..message import Message
from ..prompt_functions import PromptFunctions
from ..rendered_prompt_section import RenderedPromptSection
from .prompt_section_base import PromptSectionBase


class DataSourceSection(PromptSectionBase):
    """
    A section that renders a data source to a prompt.
    """

    _data_source: DataSource

    def __init__(self, data_source: DataSource, tokens: float):
        """
        Creates a new `DataSourceSection` instance.

        Args:
            dataSource (DataSource): The data source to render.
            tokens (int): Sizing strategy for this section.
        """
        super().__init__(tokens, True, "\n\n")
        self._data_source = data_source

    async def render_as_messages(
        self,
        context: TurnContext,
        memory: MemoryBase,
        functions: PromptFunctions,
        tokenizer: Tokenizer,
        max_tokens: int,
    ) -> RenderedPromptSection[List[Message[str]]]:
        """
        Renders the section as a list of messages.

        Args:
            context (TurnContext): Context for the current turn of conversation with the user.
            memory (MemoryBase): An interface for accessing state values.
            functions (PromptFunctions): Registry of functions that can be used by the section.
            tokenizer (Tokenizer): Tokenizer to use when rendering the section.
            max_tokens (int): Maximum number of tokens allowed to be rendered.

        Returns:
            RenderedPromptSection[List[Message]]: The rendered prompt section as a list of messages.
        """
        # Render data source
        budget = self._get_token_budget(max_tokens)
        rendered = await self._data_source.render_data(context, memory, tokenizer, budget)

        # Return as a 'system' message
        # - The role will typically end up being ignored because as this section is usually added
        #   to a `GroupSection` which will override the role.
        return RenderedPromptSection(
            output=[Message[str]("system", rendered.output)],
            length=rendered.length,
            too_long=rendered.too_long,
        )



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from typing import cast
from unittest import IsolatedAsyncioTestCase, mock

import httpx
import openai
from openai.types import chat

from teams.ai.models import AzureOpenAIModelOptions, OpenAIModel, OpenAIModelOptions
from teams.ai.prompts import (
    CompletionConfig,
    PromptFunctions,
    PromptTemplate,
    PromptTemplateConfig,
    TextSection,
)
from teams.ai.tokenizers import GPTTokenizer
from teams.state import TurnState


class MockAsyncCompletions:
    should_error = False

    def __init__(self, should_error=False) -> None:
        self.should_error = should_error

    async def create(self, **kwargs) -> chat.ChatCompletion:
        if self.should_error:
            raise openai.BadRequestError(
                "bad request",
                response=httpx.Response(400, request=httpx.Request(method="method", url="url")),
                body=None,
            )

        return chat.ChatCompletion(
            id="",
            choices=[
                chat.chat_completion.Choice(
                    finish_reason="stop",
                    index=0,
                    message=chat.ChatCompletionMessage(content="test", role="assistant"),
                )
            ],
            created=0,
            model=kwargs["model"],
            object="chat.completion",
        )


class MockAsyncChat:
    completions: MockAsyncCompletions

    def __init__(self, should_error=False) -> None:
        self.completions = MockAsyncCompletions(should_error=should_error)


class MockAsyncOpenAI:
    chat = MockAsyncChat()


class MockAsyncAzureOpenAI:
    chat = MockAsyncChat()


class MockAsyncOpenAIError:
    chat = MockAsyncChat(should_error=True)


class TestOpenAIModel(IsolatedAsyncioTestCase):
    def create_mock_context(
        self, channel_id="channel1", bot_id="bot1", conversation_id="conversation1", user_id="user1"
    ):
        context = mock.MagicMock()
        context.activity.channel_id = channel_id
        context.activity.recipient.id = bot_id
        context.activity.conversation.id = conversation_id
        context.activity.from_property.id = user_id
        return context

    @mock.patch("openai.AsyncOpenAI", spec=MockAsyncOpenAI)
    async def test_should_be_openai(self, mock_async_openai):
        model = OpenAIModel(OpenAIModelOptions(api_key="", default_model="model"))
        self.assertTrue(mock_async_openai.called)
        self.assertTrue(isinstance(model._client, MockAsyncOpenAI))

    @mock.patch("openai.AsyncAzureOpenAI", spec=MockAsyncAzureOpenAI)
    async def test_should_be_azure_openai(self, mock_async_openai):
        model = OpenAIModel(AzureOpenAIModelOptions(api_key="", default_model="model", endpoint=""))
        self.assertTrue(mock_async_openai.called)
        self.assertTrue(isinstance(model._client, MockAsyncAzureOpenAI))

    @mock.patch("openai.AsyncOpenAI", return_value=MockAsyncOpenAIError)
    async def test_should_raise_error(self, mock_async_openai):
        context = self.create_mock_context()
        state = TurnState()
        await state.load(context)

        model = OpenAIModel(OpenAIModelOptions(api_key="", default_model="model"))
        res = await model.complete_prompt(
            context=context,
            memory=state,
            functions=cast(PromptFunctions, {}),
            tokenizer=GPTTokenizer(),
            template=PromptTemplate(
                name="default",
                prompt=TextSection(text="this is a test prompt", role="system", tokens=1),
                config=PromptTemplateConfig(
                    schema=1.0,
                    type="completion",
                    description="test",
                    completion=CompletionConfig(completion_type="chat"),
                ),
            ),
        )

        self.assertTrue(mock_async_openai.called)
        self.assertEqual(res.status, "error")

    @mock.patch("openai.AsyncOpenAI", return_value=MockAsyncOpenAI)
    async def test_should_be_success(self, mock_async_openai):
        context = self.create_mock_context()
        state = TurnState()
        await state.load(context)

        model = OpenAIModel(OpenAIModelOptions(api_key="", default_model="model"))
        res = await model.complete_prompt(
            context=context,
            memory=state,
            functions=cast(PromptFunctions, {}),
            tokenizer=GPTTokenizer(),
            template=PromptTemplate(
                name="default",
                prompt=TextSection(text="this is a test prompt", role="system", tokens=1),
                config=PromptTemplateConfig(
                    schema=1.0,
                    type="completion",
                    description="test",
                    completion=CompletionConfig(completion_type="chat"),
                ),
            ),
        )

        self.assertTrue(mock_async_openai.called)
        self.assertEqual(res.status, "success")



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from unittest import TestCase

from teams.ai.prompts import AssistantMessage


class TestAssistantMessage(TestCase):
    def test_init_default_values(self):
        assistant_message = AssistantMessage("template")
        self.assertEqual(assistant_message.template, "template")
        self.assertEqual(assistant_message.tokens, -1)
        self.assertEqual(assistant_message.text_prefix, "assistant: ")
        self.assertEqual(assistant_message.role, "assistant")
        self.assertTrue(assistant_message.required)
        self.assertEqual(assistant_message.separator, "\n")

    def test_init_with_optional_params(self):
        assistant_message = AssistantMessage("template", 1, "text_prefix")
        self.assertEqual(assistant_message.template, "template")
        self.assertEqual(assistant_message.tokens, 1)
        self.assertEqual(assistant_message.text_prefix, "text_prefix")
        self.assertEqual(assistant_message.role, "assistant")
        self.assertTrue(assistant_message.required)
        self.assertEqual(assistant_message.separator, "\n")



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Generic, Literal, Optional, TypeVar

from ..prompts.message import Message

ContentT = TypeVar("ContentT")
PromptResponseStatus = Literal["success", "error", "rate_limited", "invalid_response", "too_long"]


@dataclass
class PromptResponse(Generic[ContentT]):
    """
    Response returned by a `PromptCompletionClient`.
    """

    status: PromptResponseStatus = "success"
    """
    Status of the prompt response.
    """

    input: Optional[Message[Any]] = None
    """
    User input message sent to the model. `undefined` if no input was sent.
    """

    message: Optional[Message[ContentT]] = None
    """
    Message returned.
    """

    error: Optional[str] = None
    """
    Error returned.
    """



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from typing import cast
from unittest import IsolatedAsyncioTestCase, mock

import httpx
import openai
from openai.types import chat

from teams.ai.clients.llm_client import LLMClient, LLMClientOptions
from teams.ai.models.openai_model import OpenAIModel, OpenAIModelOptions
from teams.ai.models.prompt_response import PromptResponse
from teams.ai.prompts import Message
from teams.ai.prompts.completion_config import CompletionConfig
from teams.ai.prompts.prompt_functions import PromptFunctions
from teams.ai.prompts.prompt_template import PromptTemplate
from teams.ai.prompts.prompt_template_config import PromptTemplateConfig
from teams.ai.prompts.sections.text_section import TextSection
from teams.ai.tokenizers.gpt_tokenizer import GPTTokenizer
from teams.state import TurnState
from teams.state.conversation_state import ConversationState
from teams.state.temp_state import TempState
from teams.state.user_state import UserState


class MockAsyncCompletions:
    should_error = False

    def __init__(self, should_error=False) -> None:
        self.should_error = should_error

    async def create(self, **kwargs) -> chat.ChatCompletion:
        if self.should_error:
            raise openai.BadRequestError(
                "bad request",
                response=httpx.Response(400, request=httpx.Request(method="method", url="url")),
                body=None,
            )

        return chat.ChatCompletion(
            id="",
            choices=[
                chat.chat_completion.Choice(
                    finish_reason="stop",
                    index=0,
                    message=chat.ChatCompletionMessage(content="test", role="assistant"),
                )
            ],
            created=0,
            model=kwargs["model"],
            object="chat.completion",
        )


class MockAsyncChat:
    completions: MockAsyncCompletions

    def __init__(self, should_error=False) -> None:
        self.completions = MockAsyncCompletions(should_error=should_error)


class MockAsyncOpenAI:
    chat = MockAsyncChat()


class MockAsyncOpenAIError:
    chat = MockAsyncChat(should_error=True)


class TestLLMClient(IsolatedAsyncioTestCase):
    def create_mock_context(
        self, channel_id="channel1", bot_id="bot1", conversation_id="conversation1", user_id="user1"
    ):
        context = mock.MagicMock()
        context.activity.channel_id = channel_id
        context.activity.recipient.id = bot_id
        context.activity.conversation.id = conversation_id
        context.activity.from_property.id = user_id
        return context

    @mock.patch("openai.AsyncOpenAI", return_value=MockAsyncOpenAI)
    async def test_add_function_result_object(self, mock_async_openai):
        context = self.create_mock_context()
        state = await TurnState[ConversationState, UserState, TempState].load(context)

        model = OpenAIModel(OpenAIModelOptions(api_key="", default_model="model"))
        client = LLMClient(LLMClientOptions(model))
        client.add_function_result_to_history(memory=state, name="name", results="results")

        expected_history = [Message(role="function", name="name", content='"results"')]

        self.assertTrue(mock_async_openai.called)
        self.assertEqual(state.get(client.options.history_variable), expected_history)

    @mock.patch("openai.AsyncOpenAI", return_value=MockAsyncOpenAI)
    async def test_add_function_result_bool(self, mock_async_openai):
        context = self.create_mock_context()
        state = await TurnState[ConversationState, UserState, TempState].load(context)

        model = OpenAIModel(OpenAIModelOptions(api_key="", default_model="model"))
        client = LLMClient(LLMClientOptions(model))
        client.add_function_result_to_history(memory=state, name="name", results=True)

        expected_history = [Message(role="function", name="name", content="true")]

        self.assertTrue(mock_async_openai.called)
        self.assertEqual(state.get(client.options.history_variable), expected_history)

    @mock.patch("openai.AsyncOpenAI", return_value=MockAsyncOpenAI)
    async def test_complete_prompt_no_attempts(self, mock_async_openai):
        context = self.create_mock_context()
        state = await TurnState[ConversationState, UserState, TempState].load(context)

        model = OpenAIModel(OpenAIModelOptions(api_key="", default_model="model"))
        client = LLMClient(LLMClientOptions(model))
        response = await client.complete_prompt(
            context=context,
            memory=state,
            functions=cast(PromptFunctions, {}),
            tokenizer=GPTTokenizer(),
            template=PromptTemplate(
                name="default",
                prompt=TextSection(text="this is a test prompt", role="system", tokens=1),
                config=PromptTemplateConfig(
                    schema=1.0,
                    type="completion",
                    description="test",
                    completion=CompletionConfig(completion_type="chat"),
                ),
            ),
            remaining_attempts=-2,
        )

        expected_response = PromptResponse(
            status="invalid_response", error="Reached max model response repair attempts."
        )

        self.assertTrue(mock_async_openai.called)
        self.assertEqual(response, expected_response)

    @mock.patch("openai.AsyncOpenAI", return_value=MockAsyncOpenAIError)
    async def test_complete_prompt_unsuccessful(self, mock_async_openai):
        context = self.create_mock_context()
        state = await TurnState[ConversationState, UserState, TempState].load(context)

        model = OpenAIModel(OpenAIModelOptions(api_key="", default_model="model"))
        client = LLMClient(LLMClientOptions(model))
        response = await client.complete_prompt(
            context=context,
            memory=state,
            functions=cast(PromptFunctions, {}),
            tokenizer=GPTTokenizer(),
            template=PromptTemplate(
                name="default",
                prompt=TextSection(text="this is a test prompt", role="system", tokens=1),
                config=PromptTemplateConfig(
                    schema=1.0,
                    type="completion",
                    description="test",
                    completion=CompletionConfig(completion_type="chat"),
                ),
            ),
        )

        self.assertTrue(mock_async_openai.called)
        self.assertEqual(response.status, "error")

    @mock.patch("openai.AsyncOpenAI", return_value=MockAsyncOpenAI)
    async def test_complete_prompt_with_value(self, mock_async_openai):
        context = self.create_mock_context()
        state = await TurnState[ConversationState, UserState, TempState].load(context)

        model = OpenAIModel(OpenAIModelOptions(api_key="", default_model="model"))
        client = LLMClient(LLMClientOptions(model))
        response = await client.complete_prompt(
            context=context,
            memory=state,
            functions=cast(PromptFunctions, {}),
            tokenizer=GPTTokenizer(),
            template=PromptTemplate(
                name="default",
                prompt=TextSection(text="this is a test prompt", role="system", tokens=1),
                config=PromptTemplateConfig(
                    schema=1.0,
                    type="completion",
                    description="test",
                    completion=CompletionConfig(completion_type="chat"),
                ),
            ),
            remaining_attempts=2,
        )

        expected_history = [
            Message(role="user", content=""),
            Message(role="assistant", content="test"),
        ]

        self.assertTrue(mock_async_openai.called)

        if response.message is not None:
            self.assertEqual(response.message.content, "test")

        self.assertEqual(state.get(client.options.history_variable), expected_history)



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from typing import Dict, List, Optional

from botbuilder.core import Storage, TurnContext

from ..input_file import InputFile
from .state import State, state


@state
class TempState(State):
    """
    Default Temp State
    """

    input: str = ""
    "Input passed from the user to the AI Library"

    input_files: List[InputFile] = []
    "Downloaded files passed by the user to the AI Library"

    last_output: str = ""
    "Output returned from the last executed action"

    action_outputs: Dict[str, str] = {}
    "All outputs returned from the action sequence that was executed"

    auth_tokens: Dict[str, str] = {}
    "User authentication tokens"

    duplicate_token_exchange: Optional[bool] = None
    "Flag indicating whether a token exchange event has already been processed"

    async def save(self, _context: TurnContext, storage: Optional[Storage] = None) -> None:
        return

    @classmethod
    async def load(cls, context: TurnContext, storage: Optional[Storage] = None) -> "TempState":
        # pylint: disable=unused-argument
        return cls()



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from unittest import IsolatedAsyncioTestCase
from unittest.mock import MagicMock

from botbuilder.core import TurnContext

from teams.ai.prompts import FunctionResponseMessage, PromptFunctions
from teams.ai.tokenizers import GPTTokenizer
from teams.state import Memory


class TestFunctionResponseMessage(IsolatedAsyncioTestCase):
    def setUp(self) -> None:
        self.memory = MagicMock(spec=Memory)
        self.prompt_functions = MagicMock(spec=PromptFunctions)
        self.turn_context = MagicMock(spec=TurnContext)

    def test_init(self):
        function_response_message = FunctionResponseMessage("foo", "bar")
        self.assertEqual(function_response_message.name, "foo")
        self.assertEqual(function_response_message.response, "bar")
        self.assertEqual(function_response_message.tokens, -1)
        self.assertTrue(function_response_message.required)
        self.assertEqual(function_response_message.separator, "\n")
        self.assertEqual(function_response_message.text_prefix, "user: ")

    def test_init_with_optional_params(self):
        function_response_message = FunctionResponseMessage("foo", "bar", 100, "text_prefix")
        self.assertEqual(function_response_message.name, "foo")
        self.assertEqual(function_response_message.response, "bar")
        self.assertEqual(function_response_message.tokens, 100)
        self.assertTrue(function_response_message.required)
        self.assertEqual(function_response_message.separator, "\n")
        self.assertEqual(function_response_message.text_prefix, "text_prefix")

    async def test_render_as_messages(self):
        function_response_message = FunctionResponseMessage("foo", "bar")
        rendered = await function_response_message.render_as_messages(
            self.turn_context, self.memory, self.prompt_functions, GPTTokenizer(), 100
        )
        self.assertEqual(rendered.output[0].role, "function")
        self.assertEqual(rendered.output[0].content, '"bar"')
        self.assertEqual(rendered.output[0].name, "foo")
        self.assertEqual(rendered.length, 4)
        self.assertFalse(rendered.too_long)



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from unittest import IsolatedAsyncioTestCase
from unittest.mock import MagicMock

from botbuilder.core import TurnContext

from teams.ai.prompts import GroupSection, PromptFunctions, TextSection
from teams.ai.tokenizers import GPTTokenizer
from teams.state import Memory


class TestGroupSection(IsolatedAsyncioTestCase):
    def setUp(self) -> None:
        self.memory = MagicMock(spec=Memory)
        self.prompt_functions = MagicMock(spec=PromptFunctions)
        self.turn_context = MagicMock(spec=TurnContext)

    def test_init(self):
        group_section = GroupSection([TextSection("foo", "user"), TextSection("bar", "user")])
        self.assertEqual(len(group_section.sections), 2)
        self.assertEqual(group_section.role, "system")
        self.assertEqual(group_section.tokens, -1)
        self.assertTrue(group_section.required)
        self.assertEqual(group_section.separator, "\n\n")
        self.assertEqual(group_section.text_prefix, "")

    def test_init_with_optional_params(self):
        group_section = GroupSection(
            [TextSection("foo", "user"), TextSection("bar", "user")],
            "role",
            100,
            False,
            "separator",
            "text_prefix",
        )
        self.assertEqual(len(group_section.sections), 2)
        self.assertEqual(group_section.role, "role")
        self.assertEqual(group_section.tokens, 100)
        self.assertFalse(group_section.required)
        self.assertEqual(group_section.separator, "separator")
        self.assertEqual(group_section.text_prefix, "text_prefix")

    async def test_render_as_messages(self):
        group_section = GroupSection([TextSection("foo", "user"), TextSection("bar", "user")])
        rendered = await group_section.render_as_messages(
            self.turn_context, self.memory, self.prompt_functions, GPTTokenizer(), 100
        )
        self.assertEqual(len(rendered.output), 1)
        self.assertEqual(rendered.output[0].role, "system")
        self.assertEqual(rendered.output[0].content, "foo\n\nbar")
        self.assertEqual(rendered.length, 3)
        self.assertFalse(rendered.too_long)

    async def test_render_as_messages_too_long(self):
        group_section = GroupSection([TextSection("foo", "user"), TextSection("bar", "user")])
        rendered = await group_section.render_as_messages(
            self.turn_context, self.memory, self.prompt_functions, GPTTokenizer(), 1
        )
        self.assertEqual(len(rendered.output), 1)
        self.assertEqual(rendered.output[0].role, "system")
        self.assertEqual(rendered.output[0].content, "foo\n\nbar")
        self.assertEqual(rendered.length, 3)
        self.assertTrue(rendered.too_long)



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

import asyncio
from datetime import datetime
from logging import Logger
from operator import attrgetter
from typing import List, Union

import openai

from teams.ai.embeddings.embeddings_model import EmbeddingsModel
from teams.ai.embeddings.embeddings_response import EmbeddingsResponse
from teams.ai.embeddings.openai_embeddings_options import OpenAIEmbeddingsOptions


class OpenAIEmbeddings(EmbeddingsModel):
    """
    A `EmbeddingsModel` for calling the OpenAI hosted model.
    """

    _log: Logger

    options: OpenAIEmbeddingsOptions
    "Options the client was configured with."

    def __init__(self, options: OpenAIEmbeddingsOptions, log=Logger("teams.ai")) -> None:
        """
        Creates a new `OpenAIEmbeddings` instance.

        Args:
            options (OpenAIEmbeddingsOptions): Options for configuring the embeddings client.
            log (Logger): Logger to use.
        """

        self.options = options
        self._log = log
        if not self.options.retry_policy:
            self.options.retry_policy = [2, 5]

    async def create_embeddings(
        self, inputs: Union[str, List[str], List[int], List[List[int]]], retry_count=0
    ) -> EmbeddingsResponse:
        """
        Creates embeddings for the given inputs.

        Args:
            inputs(Union[str, List[str]]): Text inputs to create embeddings for.

        Returns:
            EmbeddingsResponse: A status and embeddings/message when an error occurs.
        """

        if self.options.log_requests:
            self._log.info("Embeddings REQUEST: inputs=%s", inputs)

        if not self.options.request_config:
            self.options.request_config = {"Authorization": f"Bearer {self.options.api_key}"}
        else:
            self.options.request_config.update({"Authorization": f"Bearer {self.options.api_key}"})

        if not self.options.request_config.get("Content-Type"):
            self.options.request_config.update({"Content-Type": "application/json"})

        if not self.options.request_config.get("User-Agent"):
            self.options.request_config.update({"User-Agent": self.user_agent})

        if self.options.organization:
            self.options.request_config.update({"OpenAI-Organization": self.options.organization})

        client = openai.AsyncOpenAI(
            api_key=self.options.api_key, default_headers=self.options.request_config
        )

        try:
            start_time = datetime.now()
            res = await client.embeddings.create(input=inputs, model=self.options.model)

            data = list(map(attrgetter("embedding"), sorted(res.data, key=lambda x: x.index)))

            if self.options.log_requests:
                duration = datetime.now() - start_time
                self._log.info(
                    "Embeddings SUCCEEDED: duration=%s response=%s", duration.total_seconds, data
                )

            return EmbeddingsResponse(status="success", output=data)
        except openai.RateLimitError:
            if self.options.retry_policy:
                if retry_count < len(self.options.retry_policy):
                    delay = self.options.retry_policy[retry_count]
                    await asyncio.sleep(delay)
                    return await self.create_embeddings(inputs, retry_count + 1)
            return EmbeddingsResponse(
                status="rate_limited", output="The embeddings API returned a rate limit error."
            )
        except openai.APIError as err:
            return EmbeddingsResponse(
                status="error",
                output=f"The embeddings API returned an error status of {err.code}: {err.message}",
            )



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

# pylint:disable=duplicate-code

from unittest import IsolatedAsyncioTestCase, mock

import pytest
from botbuilder.core import TurnContext
from botbuilder.schema import Activity, ChannelAccount, ConversationAccount

from teams import Application
from teams.message_reaction_types import MessageReactionTypes
from tests.utils import SimpleAdapter


class TestApp(IsolatedAsyncioTestCase):
    app: Application

    @pytest.fixture(autouse=True)
    def before_each(self):
        self.app = Application()
        yield

    @pytest.mark.asyncio
    async def test_activity(self):
        on_event = mock.AsyncMock()
        self.app.activity("event")(on_event)
        self.assertEqual(len(self.app._routes), 1)

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="event",
                    text="test",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="UnitTest",
                    locale="en-uS",
                    service_url="https://example.org",
                ),
            )
        )

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="message",
                    text="test",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="UnitTest",
                    locale="en-uS",
                    service_url="https://example.org",
                ),
            )
        )

        on_event.assert_called_once()

    @pytest.mark.asyncio
    async def test_message(self):
        on_message = mock.AsyncMock()
        self.app.message("/history")(on_message)
        self.assertEqual(len(self.app._routes), 1)

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="event",
                    text="test",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="UnitTest",
                    locale="en-uS",
                    service_url="https://example.org",
                ),
            )
        )

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="message",
                    text="/history",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="UnitTest",
                    locale="en-uS",
                    service_url="https://example.org",
                ),
            )
        )

        on_message.assert_called_once()

    @pytest.mark.asyncio
    async def test_conversation_update(self):
        on_member_added = mock.AsyncMock()
        self.app.conversation_update("membersAdded")(on_member_added)
        self.assertEqual(len(self.app._routes), 1)

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="conversationUpdate",
                    text="test",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="UnitTest",
                    locale="en-uS",
                    service_url="https://example.org",
                    members_added=[ChannelAccount(id="user-2", name="User Name 2")],
                ),
            )
        )

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="conversationUpdate",
                    text="test",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="UnitTest",
                    locale="en-uS",
                    service_url="https://example.org",
                ),
            )
        )

        on_member_added.assert_called_once()

    @pytest.mark.asyncio
    async def test_message_reaction(self):
        on_reactions_added = mock.AsyncMock()
        on_reactions_removed = mock.AsyncMock()
        self.app.message_reaction("reactionsAdded")(on_reactions_added)
        self.app.message_reaction("reactionsRemoved")(on_reactions_removed)
        self.assertEqual(len(self.app._routes), 2)

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="messageReaction",
                    text="test",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="UnitTest",
                    locale="en-uS",
                    service_url="https://example.org",
                    reactions_added=[{type: MessageReactionTypes.LIKE}],
                ),
            )
        )

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="messageReaction",
                    text="test",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="UnitTest",
                    locale="en-uS",
                    service_url="https://example.org",
                    reactions_removed=[{type: MessageReactionTypes.PLUS_ONE}],
                ),
            )
        )

        on_reactions_added.assert_called_once()
        on_reactions_removed.assert_called_once()

    @pytest.mark.asyncio
    async def test_message_update(self):
        on_edit_message = mock.AsyncMock()
        on_soft_delete_message = mock.AsyncMock()
        on_undelete_message = mock.AsyncMock()
        self.app.message_update("editMessage")(on_edit_message)
        self.app.message_update("softDeleteMessage")(on_soft_delete_message)
        self.app.message_update("undeleteMessage")(on_undelete_message)
        self.assertEqual(len(self.app._routes), 3)

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="messageUpdate",
                    text="test",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="UnitTest",
                    channel_data={"event_type": "editMessage"},
                    locale="en-uS",
                    service_url="https://example.org",
                ),
            )
        )

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="messageDelete",
                    text="test",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="UnitTest",
                    channel_data={"event_type": "softDeleteMessage"},
                    locale="en-uS",
                    service_url="https://example.org",
                ),
            )
        )

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="messageUpdate",
                    text="test",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="UnitTest",
                    channel_data={"event_type": "undeleteMessage"},
                    locale="en-uS",
                    service_url="https://example.org",
                ),
            )
        )

        on_edit_message.assert_called_once()
        on_soft_delete_message.assert_called_once()
        on_undelete_message.assert_called_once()

    @pytest.mark.asyncio
    async def test_o365_connector_card_action(self):
        on_o365_connector_card_action = mock.AsyncMock()
        self.app.o365_connector_card_action()(on_o365_connector_card_action)

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="invoke",
                    name="actionableMessage/executeAction",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="UnitTest",
                    locale="en-uS",
                    service_url="https://example.org",
                    value={"body": "some results", "action_id": "actionId"},
                ),
            )
        )

        on_o365_connector_card_action.assert_called_once()

    @pytest.mark.asyncio
    async def test_o365_connector_card_action_wrong_type(self):
        on_o365_connector_card_action = mock.AsyncMock()
        self.app.o365_connector_card_action()(on_o365_connector_card_action)

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="event",
                    name="actionableMessage/executeAction",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="UnitTest",
                    locale="en-uS",
                    service_url="https://example.org",
                    value={"body": "some results", "action_id": "actionId"},
                ),
            )
        )

        on_o365_connector_card_action.assert_not_called()

    @pytest.mark.asyncio
    async def test_o365_connector_card_action_wrong_name(self):
        on_o365_connector_card_action = mock.AsyncMock()
        self.app.o365_connector_card_action()(on_o365_connector_card_action)

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="invoke",
                    name="actionableMessage/executeAc",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="UnitTest",
                    locale="en-uS",
                    service_url="https://example.org",
                    value={"body": "some results", "action_id": "actionId"},
                ),
            )
        )

        on_o365_connector_card_action.assert_not_called()

    @pytest.mark.asyncio
    async def test_o365_connector_card_action_missing_value(self):
        on_o365_connector_card_action = mock.AsyncMock()
        self.app.o365_connector_card_action()(on_o365_connector_card_action)

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="invoke",
                    name="actionableMessage/executeAction",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="UnitTest",
                    locale="en-uS",
                    service_url="https://example.org",
                ),
            )
        )

        on_o365_connector_card_action.assert_not_called()

    @pytest.mark.asyncio
    async def test_file_consent_accept(self):
        on_file_consent_accept = mock.AsyncMock()
        self.app.file_consent_accept()(on_file_consent_accept)

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="invoke",
                    name="fileConsent/invoke",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="UnitTest",
                    locale="en-uS",
                    service_url="https://example.org",
                    value={
                        "action": "accept",
                        "context": {"theme": "dark", "consentId": "1234567890"},
                        "uploadInfo": {
                            "name": "test.txt",
                            "uploadUrl": "https://test.com",
                            "contentUrl": "https://test.com",
                            "uniqueId": "1234567890",
                        },
                    },
                ),
            )
        )

        on_file_consent_accept.assert_called_once()

    @pytest.mark.asyncio
    async def test_file_consent_accept_wrong_type(self):
        on_file_consent_accept = mock.AsyncMock()
        self.app.file_consent_accept()(on_file_consent_accept)

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="event",
                    name="fileConsent/invoke",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="UnitTest",
                    locale="en-uS",
                    service_url="https://example.org",
                    value={
                        "action": "accept",
                        "context": {"theme": "dark", "consentId": "1234567890"},
                        "uploadInfo": {
                            "name": "test.txt",
                            "uploadUrl": "https://test.com",
                            "contentUrl": "https://test.com",
                            "uniqueId": "1234567890",
                        },
                    },
                ),
            )
        )

        on_file_consent_accept.assert_not_called()

    @pytest.mark.asyncio
    async def test_file_consent_accept_invalid_name(self):
        on_file_consent_accept = mock.AsyncMock()
        self.app.file_consent_accept()(on_file_consent_accept)

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="invoke",
                    name="fileConsent/event",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="UnitTest",
                    locale="en-uS",
                    service_url="https://example.org",
                    value={
                        "action": "accept",
                        "context": {"theme": "dark", "consentId": "1234567890"},
                        "uploadInfo": {
                            "name": "test.txt",
                            "uploadUrl": "https://test.com",
                            "contentUrl": "https://test.com",
                            "uniqueId": "1234567890",
                        },
                    },
                ),
            )
        )

        on_file_consent_accept.assert_not_called()

    @pytest.mark.asyncio
    async def test_file_consent_accept_missing_action(self):
        on_file_consent_accept = mock.AsyncMock()
        self.app.file_consent_accept()(on_file_consent_accept)

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="invoke",
                    name="fileConsent/invoke",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="UnitTest",
                    locale="en-uS",
                    service_url="https://example.org",
                    value={
                        "context": {"theme": "dark", "consentId": "1234567890"},
                        "uploadInfo": {
                            "name": "test.txt",
                            "uploadUrl": "https://test.com",
                            "contentUrl": "https://test.com",
                            "uniqueId": "1234567890",
                        },
                    },
                ),
            )
        )

        on_file_consent_accept.assert_not_called()

    @pytest.mark.asyncio
    async def test_file_consent_accept_missing_value(self):
        on_file_consent_accept = mock.AsyncMock()
        self.app.file_consent_accept()(on_file_consent_accept)

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="invoke",
                    name="fileConsent/invoke",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="UnitTest",
                    locale="en-uS",
                    service_url="https://example.org",
                ),
            )
        )

        on_file_consent_accept.assert_not_called()

    @pytest.mark.asyncio
    async def test_file_consent_accept_invalid_action(self):
        on_file_consent_accept = mock.AsyncMock()
        self.app.file_consent_accept()(on_file_consent_accept)

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="invoke",
                    name="fileConsent/invoke",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="UnitTest",
                    locale="en-uS",
                    service_url="https://example.org",
                    value={
                        "action": "stub",
                        "context": {"theme": "dark", "consentId": "1234567890"},
                        "uploadInfo": {
                            "name": "test.txt",
                            "uploadUrl": "https://test.com",
                            "contentUrl": "https://test.com",
                            "uniqueId": "1234567890",
                        },
                    },
                ),
            )
        )

        on_file_consent_accept.assert_not_called()

    @pytest.mark.asyncio
    async def test_file_consent_decline(self):
        on_file_consent_decline = mock.AsyncMock()
        self.app.file_consent_decline()(on_file_consent_decline)

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="invoke",
                    name="fileConsent/invoke",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="UnitTest",
                    locale="en-uS",
                    service_url="https://example.org",
                    value={
                        "action": "decline",
                        "context": {"theme": "dark", "consentId": "1234567890"},
                        "uploadInfo": {
                            "name": "test.txt",
                            "uploadUrl": "https://test.com",
                            "contentUrl": "https://test.com",
                            "uniqueId": "1234567890",
                        },
                    },
                ),
            )
        )

        on_file_consent_decline.assert_called_once()

    @pytest.mark.asyncio
    async def test_file_consent_decline_wrong_type(self):
        on_file_consent_decline = mock.AsyncMock()
        self.app.file_consent_decline()(on_file_consent_decline)

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="event",
                    name="fileConsent/invoke",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="UnitTest",
                    locale="en-uS",
                    service_url="https://example.org",
                    value={
                        "action": "decline",
                        "context": {"theme": "dark", "consentId": "1234567890"},
                        "uploadInfo": {
                            "name": "test.txt",
                            "uploadUrl": "https://test.com",
                            "contentUrl": "https://test.com",
                            "uniqueId": "1234567890",
                        },
                    },
                ),
            )
        )

        on_file_consent_decline.assert_not_called()

    @pytest.mark.asyncio
    async def test_file_consent_decline_invalid_name(self):
        on_file_consent_decline = mock.AsyncMock()
        self.app.file_consent_decline()(on_file_consent_decline)

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="invoke",
                    name="fileConsent/event",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="UnitTest",
                    locale="en-uS",
                    service_url="https://example.org",
                    value={
                        "action": "decline",
                        "context": {"theme": "dark", "consentId": "1234567890"},
                        "uploadInfo": {
                            "name": "test.txt",
                            "uploadUrl": "https://test.com",
                            "contentUrl": "https://test.com",
                            "uniqueId": "1234567890",
                        },
                    },
                ),
            )
        )

        on_file_consent_decline.assert_not_called()

    @pytest.mark.asyncio
    async def test_file_consent_missing_value(self):
        on_file_consent_decline = mock.AsyncMock()
        self.app.file_consent_decline()(on_file_consent_decline)

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="invoke",
                    name="fileConsent/invoke",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="UnitTest",
                    locale="en-uS",
                    service_url="https://example.org",
                ),
            )
        )

        on_file_consent_decline.assert_not_called()

    @pytest.mark.asyncio
    async def test_file_consent_decline_missing_action(self):
        on_file_consent_decline = mock.AsyncMock()
        self.app.file_consent_decline()(on_file_consent_decline)

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="invoke",
                    name="fileConsent/invoke",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="UnitTest",
                    locale="en-uS",
                    service_url="https://example.org",
                    value={
                        "context": {"theme": "dark", "consentId": "1234567890"},
                        "uploadInfo": {
                            "name": "test.txt",
                            "uploadUrl": "https://test.com",
                            "contentUrl": "https://test.com",
                            "uniqueId": "1234567890",
                        },
                    },
                ),
            )
        )

        on_file_consent_decline.assert_not_called()

    @pytest.mark.asyncio
    async def test_file_consent_decline_invalid_action(self):
        on_file_consent_decline = mock.AsyncMock()
        self.app.file_consent_decline()(on_file_consent_decline)

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="invoke",
                    name="fileConsent/invoke",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="UnitTest",
                    locale="en-uS",
                    service_url="https://example.org",
                    value={
                        "action": "stub",
                        "context": {"theme": "dark", "consentId": "1234567890"},
                        "uploadInfo": {
                            "name": "test.txt",
                            "uploadUrl": "https://test.com",
                            "contentUrl": "https://test.com",
                            "uniqueId": "1234567890",
                        },
                    },
                ),
            )
        )

        on_file_consent_decline.assert_not_called()

    @pytest.mark.asyncio
    async def test_handoff(self):
        on_handoff = mock.AsyncMock()
        self.app.handoff()(on_handoff)

        await self.app.on_turn(
            TurnContext(
                SimpleAdapter(),
                Activity(
                    id="1234",
                    type="invoke",
                    name="handoff/action",
                    from_property=ChannelAccount(id="user", name="User Name"),
                    recipient=ChannelAccount(id="bot", name="Bot Name"),
                    conversation=ConversationAccount(id="convo", name="Convo Name"),
                    channel_id="UnitTest",
                    locale="en-uS",
                    service_url="https://example.org",
                    value={"continuation": "test"},
                ),
            )
        )

        on_handoff.assert_called_once()



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from typing import Generic, TypeVar

from botbuilder.core import TurnContext

DataT = TypeVar("DataT")


class ActionTurnContext(TurnContext, Generic[DataT]):
    name: str
    """
    The action name
    """

    data: DataT
    """
    Action specific data
    """

    def __init__(self, name: str, data: DataT, context: TurnContext) -> None:
        super().__init__(context)
        context.copy_to(self)
        self._turn_state = context.turn_state
        self.name = name
        self.data = data



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from datetime import datetime
from logging import Logger
from typing import Callable, Dict, Generic, Optional, TypeVar

from botbuilder.core import TurnContext
from botbuilder.schema import Activity, ActivityTypes
from botframework.connector import Channels

from ..app_error import ApplicationError
from ..state import TurnState
from ..utils import snippet
from ..utils.citations import format_citations_response, get_used_citations
from .actions import ActionEntry, ActionHandler, ActionTurnContext, ActionTypes
from .ai_options import AIOptions
from .citations.citations import Appearance, ClientCitation
from .moderators.moderator import Moderator
from .planners.plan import Plan, PredictedDoCommand, PredictedSayCommand
from .planners.planner import Planner

StateT = TypeVar("StateT", bound=TurnState)


class AI(Generic[StateT]):
    """
    ### AI System

    The AI system is responsible for generating plans, moderating input and output, and
    generating prompts. It can be used free standing or routed to by the Application object.
    """

    _options: AIOptions
    _logger: Logger
    _actions: Dict[str, ActionEntry[StateT]] = {}

    @property
    def options(self) -> AIOptions:
        return self._options

    @property
    def planner(self) -> Planner[StateT]:
        return self._options.planner

    @property
    def moderator(self) -> Moderator[StateT]:
        return self._options.moderator

    def __init__(self, options: AIOptions, *, logger=Logger("teams.ai")) -> None:
        self._options = options
        self._logger = logger
        self._actions = {
            ActionTypes.UNKNOWN_ACTION: ActionEntry[StateT](
                ActionTypes.UNKNOWN_ACTION, True, self._on_unknown_action
            ),
            ActionTypes.FLAGGED_INPUT: ActionEntry[StateT](
                ActionTypes.FLAGGED_INPUT, True, self._on_flagged_input
            ),
            ActionTypes.FLAGGED_OUTPUT: ActionEntry[StateT](
                ActionTypes.FLAGGED_OUTPUT, True, self._on_flagged_output
            ),
            ActionTypes.HTTP_ERROR: ActionEntry[StateT](
                ActionTypes.HTTP_ERROR, True, self._on_http_error
            ),
            ActionTypes.PLAN_READY: ActionEntry[StateT](
                ActionTypes.PLAN_READY, True, self._on_plan_ready
            ),
            ActionTypes.DO_COMMAND: ActionEntry[StateT](
                ActionTypes.DO_COMMAND, True, self._on_do_command
            ),
            ActionTypes.SAY_COMMAND: ActionEntry[StateT](
                ActionTypes.SAY_COMMAND, True, self._on_say_command
            ),
            ActionTypes.TOO_MANY_STEPS: ActionEntry[StateT](
                ActionTypes.TOO_MANY_STEPS, True, self._on_too_many_steps
            ),
        }

    def action(
        self, name: Optional[str] = None, *, allow_overrides=False
    ) -> Callable[[ActionHandler[StateT]], ActionHandler[StateT]]:
        """
        Registers a new action event listener. This method can be used as either
        a decorator or a method.

        ```python
        # Use this method as a decorator
        @app.ai.action()
        async def hello_world(context: TurnContext, state: TurnState, entities: Any, name: str):
            print("hello world!")
            return True

        # Pass a function to this method
        app.ai.action()(hello_world)
        ```

        #### Args:
        - `name`: The name of the action `Default: Function Name`
        - `allow_overrides`: If it should throw an error when duplicates
        are found `Default: False`
        """

        def __call__(func: ActionHandler[StateT]) -> ActionHandler[StateT]:
            action_name = name

            if not action_name:
                action_name = func.__name__

            existing = self._actions.get(action_name)

            if existing and not existing.allow_overrides:
                raise ApplicationError(f"""
                    The AI.action() method was called with a previously
                    registered action named \"{action_name}\".
                    """)

            self._actions[action_name] = ActionEntry[StateT](action_name, allow_overrides, func)
            return func

        return __call__

    async def run(
        self,
        context: TurnContext,
        state: StateT,
        started_at: datetime = datetime.now(),
        step: int = 0,
    ) -> bool:
        """
        Calls the configured planner to generate a plan and executes the plan that is returned.
        """

        plan: Optional[Plan] = None

        if step == 0:
            plan = await self.moderator.review_input(context, state)

        if plan is None:
            if step == 0:
                plan = await self.planner.begin_task(context, state)
            else:
                plan = await self.planner.continue_task(context, state)

            plan = await self.moderator.review_output(context, state, plan)

        res = await self._actions[ActionTypes.PLAN_READY].invoke(
            context, state, plan, ActionTypes.PLAN_READY
        )

        if res == ActionTypes.STOP:
            return False

        loop = False

        for command in plan.commands:
            step += 1
            output = ""

            if isinstance(command, PredictedDoCommand):
                if command.action in self._actions:
                    output = await self._actions[ActionTypes.DO_COMMAND].invoke(
                        context, state, command, command.action
                    )
                    loop = len(output) > 0
                    state.temp.action_outputs[command.action] = output
                else:
                    output = await self._actions[ActionTypes.UNKNOWN_ACTION].invoke(
                        context, state, plan, command.action
                    )
            elif isinstance(command, PredictedSayCommand):
                loop = False
                output = await self._actions[ActionTypes.SAY_COMMAND].invoke(
                    context, state, command, ActionTypes.SAY_COMMAND
                )
            else:
                raise ApplicationError(f"unknown command of type {command.type} predicted")

            if output == ActionTypes.STOP:
                return False

            state.temp.last_output = output
            state.temp.input = output
            state.temp.input_files = []

        if loop and self._options.allow_looping:
            return await self.run(context, state, started_at, step)

        return True

    async def _on_unknown_action(
        self,
        context: ActionTurnContext,
        _state: StateT,
    ) -> str:
        self._logger.error(
            'An AI action named "%s" was predicted but no handler was registered', context.name
        )
        return ActionTypes.STOP

    async def _on_flagged_input(
        self,
        context: ActionTurnContext,
        _state: StateT,
    ) -> str:
        self._logger.error(
            "The users input has been moderated but no handler was registered for %s", context.name
        )
        return ActionTypes.STOP

    async def _on_flagged_output(
        self,
        context: ActionTurnContext,
        _state: StateT,
    ) -> str:
        self._logger.error(
            "The apps output has been moderated but no handler was registered for %s", context.name
        )
        return ActionTypes.STOP

    async def _on_http_error(
        self,
        _context: ActionTurnContext[dict],
        _state: StateT,
    ) -> str:
        status = _context.data.get("status")
        message = _context.data.get("message")
        raise ApplicationError(
            "An AI request failed because an http error occurred. "
            f"Status code:{status}. Message:{message}"
        )

    async def _on_plan_ready(
        self,
        context: ActionTurnContext[Plan],
        _state: StateT,
    ) -> str:
        return "" if len(context.data.commands) > 0 else ActionTypes.STOP

    async def _on_do_command(
        self,
        context: ActionTurnContext[PredictedDoCommand],
        state: StateT,
    ) -> str:
        action = self._actions.get(context.data.action)
        ctx = ActionTurnContext(context.data.action, context.data.parameters, context)

        if not action:
            return await self._on_unknown_action(ctx, state)

        return await action.invoke(context, state, context.data.parameters, context.name)

    async def _on_say_command(
        self,
        context: ActionTurnContext[PredictedSayCommand],
        _state: StateT,
    ) -> str:
        content = (
            context.data.response.content
            if context.data.response and context.data.response.content
            else ""
        )
        msg_context = (
            context.data.response.context
            if context.data.response and context.data.response.context
            else None
        )
        is_teams_channel = context.activity.channel_id == Channels.ms_teams

        if not content:
            return ""

        if is_teams_channel:
            content = content.replace("\n", "<br>")

        # If the response from AI includes citations,
        # those citations will be parsed and added to the SAY command.
        citations = []

        if (
            msg_context
            and isinstance(msg_context, dict)
            and "citations" in msg_context
            and len(msg_context["citations"]) > 0
        ):
            for i, citation in enumerate(msg_context["citations"]):
                citations.append(
                    ClientCitation(
                        position=f"[{i + 1}]",
                        appearance=Appearance(
                            name=citation["title"],
                            abstract=snippet(citation["abstract"], 500),
                        ),
                    )
                )

        # If there are citations, modify the content so that the sources are numbers
        # instead of [doc1], [doc2], etc.
        content_text = content if not citations else format_citations_response(content)

        # If there are citations, filter out the citations unused in content.
        referenced_citations = get_used_citations(content_text, citations)
        channel_data = {}

        if is_teams_channel:
            channel_data["feedbackLoopEnabled"] = self._options.enable_feedback_loop

        await context.send_activity(
            Activity(
                type=ActivityTypes.message,
                text=content_text,
                channel_data=channel_data,
                entities=[
                    {
                        "type": "https://schema.org/Message",
                        "@type": "Message",
                        "@context": "https://schema.org",
                        "@id": "",
                        "additionalType": ["AIGeneratedContent"],
                        **(
                            {"citation": [citation.__dict__ for citation in referenced_citations]}
                            if referenced_citations
                            else {}
                        ),
                    }
                ],
            )
        )

        return ""

    async def _on_too_many_steps(
        self,
        _context: ActionTurnContext,
        _state: StateT,
    ) -> str:
        self._logger.error("The run retrieval for the Assistants Planner has expired.")
        return ActionTypes.STOP



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Dict, List

from botbuilder.core import TurnContext

from ...state import MemoryBase
from ..models.chat_completion_action import ChatCompletionAction
from ..models.prompt_response import PromptResponse
from ..prompts.message import Message
from ..tokenizers import Tokenizer
from .json_response_validator import JSONResponseValidator
from .prompt_response_validator import PromptResponseValidator
from .validation import Validation


@dataclass
class ValidatedChatCompletionAction:
    """
    A validated action call.
    """

    name: str
    """
    Name of the action to call.
    """

    parameters: Dict[str, Any] = field(default_factory=dict)
    """
    Arguments to pass to the action.
    """


class ActionResponseValidator(PromptResponseValidator):
    """
    Default response validator that always returns true.
    """

    _actions: Dict[str, ChatCompletionAction] = {}
    _required: bool
    _noun: str

    def __init__(
        self,
        actions: List[ChatCompletionAction],
        required: bool = False,
        noun: str = "action",
    ) -> None:
        """
        Creates a new `ActionResponseValidator` instance.
        """

        super().__init__()

        self._required = required
        self._noun = noun

        for action in actions:
            self._actions[action.name] = action

    @property
    def actions(self) -> List[ChatCompletionAction]:
        """
        Gets a list of the actions configured for the validator.
        """

        arr: List[ChatCompletionAction] = []

        for _, action in self._actions.items():
            arr.append(action)

        return arr

    async def validate_response(
        self,
        context: TurnContext,
        memory: MemoryBase,
        tokenizer: Tokenizer,
        response: PromptResponse[str],
        remaining_attempts: int,
    ) -> Validation:
        message = response.message
        func = response.message.function_call if response.message is not None else None

        if message is None or func is None:
            if not self._required:
                return Validation()

            return Validation(
                valid=False,
                feedback=(
                    f"No {self._noun} was specified. Call a {self._noun} with valid arguments."
                ),
            )

        if func.name is None:
            return Validation(
                valid=False,
                feedback=f"{self._noun} name missing. Specify a valid {self._noun} name.",
            )

        if not func.name in self._actions:
            return Validation(
                valid=False,
                feedback=(
                    f'Unknown {self._noun} named "{func.name}". Specify a valid {self._noun} name.'
                ),
            )

        params: Dict[str, Any] = {}
        action = self._actions[func.name]

        if action.parameters is not None:
            validator = JSONResponseValidator(
                schema=action.parameters,
                missing_json_feedback=(
                    f"No arguments were sent with called {self._noun}. "
                    f'Call the "{func.name}" {self._noun} with required '
                    "arguments as a valid JSON object."
                ),
                error_feedback=(
                    f"The {self._noun} arguments had errors. "
                    f'Apply these fixes and call "{func.name}" {self._noun} again:'
                ),
            )

            res = await validator.validate_response(
                context=context,
                memory=memory,
                tokenizer=tokenizer,
                remaining_attempts=remaining_attempts,
                response=PromptResponse(
                    message=Message(
                        role="assistant", content="{}" if func.arguments is None else func.arguments
                    )
                ),
            )

            if not res.valid:
                return Validation(valid=False, feedback=res.feedback)

            if res.value is not None:
                params = res.value

        return Validation(
            valid=True, value=ValidatedChatCompletionAction(name=func.name, parameters=params)
        )



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from .parse import parse
from .parse_object import parse_object

__all__ = ["parse", "parse_object"]



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

import json
import os
from copy import deepcopy
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional

from botbuilder.core import TurnContext

import teams.ai.augmentations

from ...app_error import ApplicationError
from ...state import MemoryBase
from ..data_sources import DataSource
from ..models.chat_completion_action import ChatCompletionAction
from ..tokenizers import Tokenizer
from .prompt import Prompt
from .prompt_functions import PromptFunction, PromptFunctions
from .prompt_manager_options import PromptManagerOptions
from .prompt_template import PromptTemplate
from .prompt_template_config import PromptTemplateConfig
from .sections import (
    ConversationHistorySection,
    GroupSection,
    PromptSection,
    TemplateSection,
)
from .sections.data_source_section import DataSourceSection
from .user_input_message import UserInputMessage
from .user_message import UserMessage


class PromptManager(PromptFunctions):
    _options: PromptManagerOptions
    _data_sources: Dict[str, DataSource]
    _functions: Dict[str, PromptFunction]
    _prompts: Dict[str, PromptTemplate]

    def __init__(self, options: PromptManagerOptions):
        """
        Creates a new 'PromptManager' instance.

        Args:
            options (PromptManagerOptions): Options used to configure the prompt manager.
        """
        self._options = options
        self._data_sources = {}
        self._functions = {}
        self._prompts = {}

    @property
    def options(self) -> PromptManagerOptions:
        """
        Gets the configured prompt manager options.
        """
        return self._options

    def function(self, name: Optional[str] = None) -> Callable[[PromptFunction], PromptFunction]:
        """
        Registers a new prompt function event listener. This method can be used as either
        a decorator or a method.

        ```python
        # Use this method as a decorator
        @prompts.function()
        async def hello_world(context: TurnContext, state: TurnState, entities: Any, name: str):
            print("hello world!")
            return True

        # Pass a function to this method
        prompts.function()(hello_world)
        ```

        Args:
        - `name`: The name of the action `Default: Function Name`
        """

        def __call__(func: PromptFunction) -> PromptFunction:
            func_name = name

            if not func_name:
                func_name = func.__name__

            self.add_function(func_name, func)
            return func

        return __call__

    def add_data_source(self, data_source: DataSource) -> "PromptManager":
        """
        Registers a new data source with the prompt manager.

        Args:
            data_source (DataSource): Data source to add.

        Returns:
            PromptManager: The prompt manager for chaining.

        Raises:
            ApplicationError: If a data source with the same name already exists.
        """
        if data_source.name in self._data_sources:
            raise ApplicationError(f"DataSource '{data_source.name}' already exists.")
        self._data_sources[data_source.name] = data_source
        return self

    def get_data_source(self, name: str) -> DataSource:
        """
        Looks up a data source by name.

        Args:
            name (str): Name of the data source to lookup.

        Returns:
            DataSource: The data source.

        Raises:
            ApplicationError: If the data source is not found.
        """
        if name not in self._data_sources:
            raise ApplicationError(f"DataSource '{name}' not found.")
        return self._data_sources[name]

    def has_data_source(self, name: str) -> bool:
        """
        Checks for the existence of a named data source.

        Args:
            name (str): Name of the data source to lookup.

        Returns:
            bool: True if the data source exists, False otherwise.
        """
        return name in self._data_sources

    def add_function(self, name: str, function: PromptFunction) -> "PromptManager":
        """
        Registers a new prompt template function with the prompt manager.

        Args:
            name (str): Name of the function to add.
            fn (PromptFunction): Function to add.

        Returns:
            PromptManager: The prompt manager for chaining.

        Raises:
            ApplicationError: If a function with the same name already exists.
        """
        if name in self._functions:
            raise ApplicationError(f"Function '{name}' already exists.")
        self._functions[name] = function
        return self

    def get_function(self, name: str) -> PromptFunction:
        """
        Looks up a prompt template function by name.

        Args:
            name (str): Name of the function to lookup.

        Returns:
            PromptFunction: The function.

        Raises:
            ApplicationError: If the function is not found.
        """
        if name not in self._functions:
            raise ApplicationError(f"Function '{name}' not found.")
        return self._functions[name]

    def has_function(self, name: str) -> bool:
        """
        Checks for the existence of a named prompt template function.

        Args:
            name (str): Name of the function to lookup.

        Returns:
            bool: True if the function exists, False otherwise.
        """
        return name in self._functions

    async def invoke_function(
        self,
        name: str,
        context: TurnContext,
        memory: MemoryBase,
        tokenizer: Tokenizer,
        args: List[str],
    ) -> Any:
        """
        Invokes a prompt template function by name.

        Args:
            name (str): Name of the function to invoke.
            context (TurnContext): Turn context for the current turn of conversation with the user.
            memory (MemoryBase): An interface for accessing state values.
            tokenizer (Tokenizer): Tokenizer to use when rendering the prompt.
            args (List[str]): Arguments to pass to the function.

        Returns:
            Any: Value returned by the function.
        """
        function = self.get_function(name)
        return await function(context, memory, self, tokenizer, args)

    def add_prompt(self, prompt: PromptTemplate) -> "PromptManager":
        """
        Registers a new prompt template with the prompt manager.

        Args:
            prompt (PromptTemplate): Prompt template to add.

        Returns:
            PromptManager: The prompt manager for chaining.

        Raises:
            ApplicationError: If a prompt with the same name already exists.
        """
        if prompt.name in self._prompts:
            raise ApplicationError(
                "The PromptManager.add_prompt() method was called with a "
                f"previously registered prompt named '{prompt.name}'."
            )

        # Clone and cache prompt
        self._prompts[prompt.name] = deepcopy(prompt)
        return self

    async def get_prompt(self, name: str) -> PromptTemplate:
        """
        Loads a named prompt template from the filesystem.

        The template will be pre-parsed and cached for use when the template is rendered by name.

        Any augmentations will also be added to the template.

        Args:
            name (str): Name of the prompt to load.

        Returns:
            PromptTemplate: The loaded and parsed prompt template.

        Raises:
            ApplicationError: If the prompt is not found or there is an error loading it.
        """
        if name not in self._prompts:
            template_name = name

            # Load template from disk
            folder = os.path.join(self._options.prompts_folder, name)
            config_file = os.path.join(folder, "config.json")
            prompt_file = os.path.join(folder, "skprompt.txt")
            actions_file = os.path.join(folder, "actions.json")

            # Load prompt config
            try:
                with open(config_file, "r", encoding="utf-8") as file:
                    template_config = PromptTemplateConfig.from_dict(json.load(file))
            except Exception as e:
                raise ApplicationError(
                    "PromptManager.get_prompt(): an error occurred while loading "
                    f"'{config_file}'. The file is either invalid or missing."
                ) from e

            # Load prompt text
            sections: List[PromptSection] = []
            try:
                with open(prompt_file, "r", encoding="utf-8") as file:
                    prompt = file.read()
                    sections.append(TemplateSection(prompt, self._options.role))
            except Exception as e:
                raise ApplicationError(
                    "PromptManager.get_prompt(): an error occurred while loading "
                    f"'{prompt_file}'. The file is either invalid or missing."
                ) from e

            # Load optional actions
            template_actions: List[ChatCompletionAction] = []
            try:
                with open(actions_file, "r", encoding="utf-8") as file:
                    actions = json.load(file)

                    for action in actions:
                        template_actions.append(ChatCompletionAction.from_dict(action))
            except IOError:
                # Ignore missing actions file
                pass

            # Migrate the templates config as needed
            self._update_config(template_config)

            # Add augmentations
            augmentation = self._append_augmentations(
                name, template_config, template_actions, sections
            )

            # Group everything into a system message
            sections = [GroupSection(sections, "system")]

            # Include conversation history
            # - The ConversationHistory section will use the remaining tokens from
            #   max_input_tokens.
            if template_config.completion.include_history:
                sections.append(
                    ConversationHistorySection(
                        f"conversation.{template_name}_history",
                        self._options.max_conversation_history_tokens,
                    )
                )

            # Include user input
            if template_config.completion.include_images:
                sections.append(UserInputMessage(self._options.max_input_tokens))
            elif template_config.completion.include_input:
                sections.append(UserMessage("{{$temp.input}}", self._options.max_input_tokens))

            template = PromptTemplate(
                template_name, Prompt(sections), template_config, template_actions
            )

            if augmentation:
                template.augmentation = augmentation

            # Cache loaded template
            self._prompts[name] = template

        return self._prompts[name]

    def has_prompt(self, name: str) -> bool:
        """
        Checks for the existence of a named prompt.

        Args:
            name (str): Name of the prompt to check.

        Returns:
            bool: True if the prompt exists, False otherwise.
        """
        if name not in self._prompts:
            folder = os.path.join(self._options.prompts_folder, name)
            prompt_file = os.path.join(folder, "skprompt.txt")

            return Path(prompt_file).exists()
        return True

    def _update_config(self, template_config: PromptTemplateConfig):
        # Migrate old schema
        if template_config.schema == 1:
            template_config.schema = 1.1
            if (
                template_config.default_backends is not None
                and len(template_config.default_backends) > 0
            ):
                template_config.completion.model = template_config.default_backends[0]

    def _append_augmentations(
        self,
        name: str,
        template_config: PromptTemplateConfig,
        template_actions: List[ChatCompletionAction],
        sections: List[PromptSection],
    ) -> Optional[teams.ai.augmentations.Augmentation]:
        # Check for augmentation
        augmentation = template_config.augmentation
        if augmentation:
            # First append data sources
            # - We're using a minimum of 2 tokens for each data source to prevent
            # any sort of prompt rendering conflicts between sources and conversation history.
            # - If we wanted to let users specify a percentage% for a data source we would need
            # to track the percentage they gave the data source(s) and give the remaining to
            # the ConversationHistory section.
            data_sources = augmentation.data_sources if augmentation.data_sources else {}
            for source in data_sources:
                if not self.has_data_source(source):
                    raise ApplicationError(f"DataSource '{source}' not found for prompt '{name}.")
                data_source = self.get_data_source(source)
                tokens = max(data_sources[source], 2)
                sections.append(DataSourceSection(data_source, tokens))

            # Next, create augmentation
            augmentation_type = augmentation.augmentation_type
            curr_augmentation: Optional[teams.ai.augmentations.Augmentation] = None

            # Parse the dict objects into ChatCompletionAction objects
            parsed_actions: List[ChatCompletionAction] = []
            if template_actions:
                for action in template_actions:
                    parsed_actions.append(ChatCompletionAction.from_dict(action.to_dict()))

            curr_actions = parsed_actions if template_actions else []
            if augmentation_type == "monologue":
                curr_augmentation = teams.ai.augmentations.MonologueAugmentation(curr_actions)
            elif augmentation_type == "sequence":
                curr_augmentation = teams.ai.augmentations.SequenceAugmentation(curr_actions)

            # Append the augmentations prompt section
            if curr_augmentation:
                section = curr_augmentation.create_prompt_section()
                if section:
                    sections.append(section)

            return curr_augmentation
        return None



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from unittest import IsolatedAsyncioTestCase
from unittest.mock import MagicMock

from botbuilder.core import MemoryStorage, StoreItem

from teams.state import UserState


class TestUserState(IsolatedAsyncioTestCase):
    def create_mock_context(
        self, channel_id="channel1", bot_id="bot1", conversation_id="conversation1", user_id="user1"
    ):
        context = MagicMock()
        context.activity.channel_id = channel_id
        context.activity.recipient.id = bot_id
        context.activity.conversation.id = conversation_id
        context.activity.from_property.id = user_id
        return context

    async def test_should_load(self):
        context = self.create_mock_context()
        storage = MemoryStorage()

        await storage.write(
            {"channel1/bot1/users/user1": StoreItem(hello="world", message="hello world")}
        )

        state = await UserState.load(context, storage)
        self.assertEqual(state.__key__, "channel1/bot1/users/user1")
        self.assertTrue("hello" in state)
        self.assertEqual(state["hello"], "world")
        self.assertTrue("message" in state)
        self.assertEqual(state["message"], "hello world")
        self.assertEqual(str(state), '{"hello": "world", "message": "hello world"}')

    async def test_should_save(self):
        context = self.create_mock_context()
        storage = MemoryStorage()
        state = await UserState.load(context, storage)

        self.assertEqual(state.__key__, "channel1/bot1/users/user1")
        self.assertFalse("hello" in state)
        self.assertFalse("message" in state)

        state["hello"] = "world"
        state["message"] = "hello world"
        await state.save(context, storage)
        state = await UserState.load(context, storage)

        self.assertEqual(state.__key__, "channel1/bot1/users/user1")
        self.assertTrue("hello" in state)
        self.assertEqual(state["hello"], "world")
        self.assertTrue("message" in state)
        self.assertEqual(state["message"], "hello world")
        self.assertEqual(str(state), '{"hello": "world", "message": "hello world"}')

    async def test_should_not_load_when_channel_missing(self):
        context = MagicMock()
        context.activity.channel_id = None
        storage = MemoryStorage()

        await storage.write(
            {"channel1/bot1/users/user1": StoreItem(hello="world", message="hello world")}
        )

        with self.assertRaises(ValueError, msg="missing activity.channel_id"):
            await UserState.load(context, storage)

    async def test_should_not_load_when_from_property_missing(self):
        context = MagicMock()
        context.activity.channel_id = "channel1"
        context.activity.from_property = None
        storage = MemoryStorage()

        await storage.write(
            {"channel1/bot1/users/user1": StoreItem(hello="world", message="hello world")}
        )

        with self.assertRaises(ValueError, msg="missing activity.from_property"):
            await UserState.load(context, storage)

    async def test_should_not_load_when_recipient_missing(self):
        context = MagicMock()
        context.activity.channel_id = "channel1"
        context.activity.from_property = MagicMock(id="user1")
        context.activity.recipient = None
        storage = MemoryStorage()

        await storage.write(
            {"channel1/bot1/users/user1": StoreItem(hello="world", message="hello world")}
        )

        with self.assertRaises(ValueError, msg="missing activity.recipient"):
            await UserState.load(context, storage)



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from .sections.template_section import TemplateSection


class UserMessage(TemplateSection):
    """
    A user message
    """

    def __init__(self, template: str, tokens: float = -1, user_prefix: str = "user: "):
        """
        Creates a new 'UserMessage' instance.

        Args:
            template (str): Template to use for this section.
            tokens (int, optional): Sizing strategy for this section. Defaults to `auto`.
            user_prefix (str, optional): Prefix to use for user messages when
                rendering as text. Defaults to `user: `.
        """
        super().__init__(template, "user", tokens, True, "\n", user_prefix)



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Generic, TypeVar

from ..state import TurnState
from .moderators.default_moderator import DefaultModerator
from .moderators.moderator import Moderator
from .planners.planner import Planner

StateT = TypeVar("StateT", bound=TurnState)


@dataclass
class AIOptions(Generic[StateT]):
    planner: Planner[StateT]
    """
    The planner to use for generating plans.
    """

    moderator: Moderator = field(default_factory=DefaultModerator)
    """
    Optional. The moderator to use for moderating input passed to
    the model and the output return by the model.
    """

    allow_looping: bool = True
    """
    Optional. If true, the AI system will allow the planner to loop.
    Default `True`
    """

    enable_feedback_loop: bool = False
    """
    Optional. If true, the AI system will enable the feedback loop in Teams that
    allows a user to give thumbs up or down to a response.
    """



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Dict, Literal, Optional


@dataclass
class AugmentationConfig:
    """
    Interface for the augmentation configuration portion of a prompt template.
    New in schema version 1.1.

    Attributes:
        augmentation_type (Literal['none','sequence', 'monologue']): Type of augmentation to use.

        data_sources (Optional[Dict[str, float]]): List of named data sources
            to augment the prompt with. For each data source, the value is the max number of
            tokens to use from the data source.
    """

    augmentation_type: Literal["none", "sequence", "monologue"] = "none"
    data_sources: Optional[Dict[str, float]] = None

    @classmethod
    def from_dict(cls, data: dict) -> "AugmentationConfig":
        return cls(
            augmentation_type=data.get("augmentation_type", "none"),
            data_sources=data.get("data_sources"),
        )



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from .teams_attachment_downloader import TeamsAttachmentDownloader
from .teams_attachment_downloader_options import TeamsAttachmentDownloaderOptions

__all__ = ["TeamsAttachmentDownloader", "TeamsAttachmentDownloaderOptions"]



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from abc import ABC, abstractmethod

from botbuilder.core import TurnContext

from teams.ai.prompts.rendered_prompt_section import RenderedPromptSection
from teams.ai.tokenizers import Tokenizer
from teams.state.memory import MemoryBase


class DataSource(ABC):
    """
    A data source that can be used to render text that's added to a prompt.
    """

    @property
    @abstractmethod
    def name(self) -> str:
        "Name of the data source."

    @abstractmethod
    async def render_data(
        self,
        turn_context: TurnContext,
        memory: MemoryBase,
        tokenizer: Tokenizer,
        max_tokens: int,
    ) -> RenderedPromptSection[str]:
        """
        Renders the data source as a string of text.

        Args:
            turn_context (TurnContext): The turn context for current turn of conversation.
            memory (MemoryBase): An interface for accessing state values.
            tokenizer (Tokenizer): Tokenizer to use when rendering the data source.
            max_tokens (int): Maximum number of tokens allowed to be rendered.

        Returns:
            RenderedPromptSection: The text to inject into the prompt as a
            `RenderedPromptSection` object.
        """



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

import json
from dataclasses import dataclass
from typing import Optional
from unittest import IsolatedAsyncioTestCase, mock

from botbuilder.schema import Attachment
from botframework.connector.auth import (
    Authentication,
    CallerIdConstants,
    GovernmentConstants,
    MicrosoftAppCredentials,
    PasswordServiceClientCredentialFactory,
)

from teams.input_file import InputFile
from teams.teams_adapter import TeamsAdapter
from teams.teams_attachment_downloader import (
    TeamsAttachmentDownloader,
    TeamsAttachmentDownloaderOptions,
)


class MockBFAppCredentials(MicrosoftAppCredentials):
    def __init__(self):
        super().__init__(app_id="botAppId", password="botAppPassword")

    def get_access_token(self, force_refresh: bool = False) -> str:
        return "authToken"


class MockCredentialsFactory(PasswordServiceClientCredentialFactory):
    def __init__(self):
        super().__init__(app_id="botAppId", password="botAppPassword")

    async def create_credentials(
        self,
        app_id: str,
        oauth_scope: str,
        login_endpoint: str,
        validate_authority: bool,
    ) -> Authentication:
        return MockBFAppCredentials()


@dataclass
class MockConfiguration:
    # pylint: disable=invalid-name
    # Need to replicate constants in configuration object
    TO_CHANNEL_FROM_BOT_LOGIN_URL: Optional[str] = ""
    TO_CHANNEL_FROM_BOT_OAUTH_SCOPE: Optional[str] = ""
    CALLER_ID: Optional[str] = ""


class MockedResponse:
    async def read(self):
        return bytes("file.png", "utf-8")


class TestTeamsAttachmentDownloader(IsolatedAsyncioTestCase):
    def create_mock_context(
        self, channel_id="channel1", bot_id="bot1", conversation_id="conversation1", user_id="user1"
    ):
        context = mock.MagicMock()
        context.activity.channel_id = channel_id
        context.activity.recipient.id = bot_id
        context.activity.conversation.id = conversation_id
        context.activity.from_property.id = user_id
        return context

    async def test_empty_attachments(self):
        context = self.create_mock_context()
        context.activity.type = "message"
        context.activity.text = "Here is the attachment"
        context.activity.attachments = []
        downloader = TeamsAttachmentDownloader(
            TeamsAttachmentDownloaderOptions(bot_app_id="botAppId", adapter=TeamsAdapter({}))
        )

        input_files = await downloader.download_files(context)

        self.assertEqual(input_files, [])

    async def test_no_attachments(self):
        context = self.create_mock_context()
        context.activity.type = "message"
        context.activity.text = "Here is the attachment"
        downloader = TeamsAttachmentDownloader(
            TeamsAttachmentDownloaderOptions(bot_app_id="botAppId", adapter=TeamsAdapter({}))
        )

        input_files = await downloader.download_files(context)

        self.assertEqual(input_files, [])

    async def test_html_attachment(self):
        attachment = Attachment(
            content_url="https://example.com/file.png", content_type="text/html", name="file.png"
        )

        context = self.create_mock_context()
        context.activity.type = "message"
        context.activity.text = "Here is the attachment"
        context.activity.attachments = [attachment]
        downloader = TeamsAttachmentDownloader(
            TeamsAttachmentDownloaderOptions(bot_app_id="botAppId", adapter=TeamsAdapter({}))
        )

        input_files = await downloader.download_files(context)

        self.assertEqual(input_files, [])

    @mock.patch("aiohttp.ClientSession.get")
    async def test_should_download_file(self, mock_get):
        response_obj = MockedResponse()
        mock_get.return_value.__aenter__.return_value = response_obj
        mocked_content = await response_obj.read()
        attachment = Attachment(
            content_url="https://example.com/file.png", content_type="image/png", name="file.png"
        )

        context = self.create_mock_context()
        context.activity.type = "message"
        context.activity.text = "Here is the attachment"
        context.activity.attachments = [attachment]
        downloader = TeamsAttachmentDownloader(
            TeamsAttachmentDownloaderOptions(bot_app_id="botAppId", adapter=TeamsAdapter({}))
        )

        input_files = await downloader.download_files(context)

        assert mock_get.call_count == 1
        self.assertEqual(
            input_files,
            [
                InputFile(
                    content=mocked_content,
                    content_type="image/png",
                    content_url="https://example.com/file.png",
                )
            ],
        )

    @mock.patch("aiohttp.ClientSession.get")
    async def test_should_download_local_file(self, mock_get):
        response_obj = MockedResponse()
        mock_get.return_value.__aenter__.return_value = response_obj
        mocked_content = await response_obj.read()
        attachment = Attachment(
            content_url="http://localhost:3978/file.png", content_type="image/png", name="file.png"
        )

        context = self.create_mock_context()
        context.activity.type = "message"
        context.activity.text = "Here is the attachment"
        context.activity.attachments = [attachment]
        downloader = TeamsAttachmentDownloader(
            TeamsAttachmentDownloaderOptions(bot_app_id="botAppId", adapter=TeamsAdapter({}))
        )

        input_files = await downloader.download_files(context)

        assert mock_get.call_count == 1
        self.assertEqual(
            input_files,
            [
                InputFile(
                    content=mocked_content,
                    content_type="image/png",
                    content_url="http://localhost:3978/file.png",
                )
            ],
        )

    async def test_buffered_attachment(self):
        attachment = Attachment(
            content=json.dumps("file.png").encode("utf-8"),
            content_type="image/png",
            name="file.png",
        )

        context = self.create_mock_context()
        context.activity.type = "message"
        context.activity.text = "Here is the attachment"
        context.activity.attachments = [attachment]
        downloader = TeamsAttachmentDownloader(
            TeamsAttachmentDownloaderOptions(bot_app_id="botAppId", adapter=TeamsAdapter({}))
        )

        input_files = await downloader.download_files(context)

        self.assertEqual(
            input_files,
            [
                InputFile(
                    content=json.dumps("file.png").encode("utf-8"),
                    content_type="image/png",
                    content_url=None,
                )
            ],
        )

    async def test_invalid_http_attachment(self):
        attachment = Attachment(
            content_url="http://example.com/file.png", content_type="image/png", name="file.png"
        )

        context = self.create_mock_context()
        context.activity.type = "message"
        context.activity.text = "Here is the attachment"
        context.activity.attachments = [attachment]
        downloader = TeamsAttachmentDownloader(
            TeamsAttachmentDownloaderOptions(bot_app_id="botAppId", adapter=TeamsAdapter({}))
        )

        input_files = await downloader.download_files(context)

        self.assertEqual(
            input_files,
            [
                InputFile(
                    content=bytes(),
                    content_type="image/png",
                    content_url="http://example.com/file.png",
                )
            ],
        )

    @mock.patch("aiohttp.ClientSession.get")
    async def test_should_update_mime_content_type(self, mock_get):
        response_obj = MockedResponse()
        mock_get.return_value.__aenter__.return_value = response_obj
        mocked_content = await response_obj.read()
        attachment = Attachment(
            content_url="https://example.com/file.png", content_type="image/*", name="file.png"
        )

        context = self.create_mock_context()
        context.activity.type = "message"
        context.activity.text = "Here is the attachment"
        context.activity.attachments = [attachment]
        downloader = TeamsAttachmentDownloader(
            TeamsAttachmentDownloaderOptions(bot_app_id="botAppId", adapter=TeamsAdapter({}))
        )

        input_files = await downloader.download_files(context)

        assert mock_get.call_count == 1
        self.assertEqual(
            input_files,
            [
                InputFile(
                    content=mocked_content,
                    content_type="image/png",
                    content_url="https://example.com/file.png",
                )
            ],
        )

    @mock.patch("aiohttp.ClientSession.get")
    async def test_should_download_file_with_enabled_auth(self, mock_get):
        response_obj = MockedResponse()
        mock_get.return_value.__aenter__.return_value = response_obj
        mocked_content = await response_obj.read()
        attachment = Attachment(
            content_url="https://example.com/file.png", content_type="image/png", name="file.png"
        )

        context = self.create_mock_context()
        context.activity.type = "message"
        context.activity.text = "Here is the attachment"
        context.activity.attachments = [attachment]
        downloader = TeamsAttachmentDownloader(
            TeamsAttachmentDownloaderOptions(
                bot_app_id="botAppId",
                adapter=TeamsAdapter({}, credentials_factory=MockCredentialsFactory()),
            )
        )

        input_files = await downloader.download_files(context)

        assert mock_get.call_count == 1
        self.assertEqual(
            input_files,
            [
                InputFile(
                    content=mocked_content,
                    content_type="image/png",
                    content_url="https://example.com/file.png",
                )
            ],
        )

    @mock.patch("aiohttp.ClientSession.get")
    async def test_should_download_file_with_enabled_auth_with_audience(self, mock_get):
        response_obj = MockedResponse()
        mock_get.return_value.__aenter__.return_value = response_obj
        mocked_content = await response_obj.read()
        attachment = Attachment(
            content_url="https://example.com/file.png", content_type="image/png", name="file.png"
        )

        context = self.create_mock_context()
        context.activity.type = "message"
        context.activity.text = "Here is the attachment"
        context.activity.attachments = [attachment]
        config = MockConfiguration(
            TO_CHANNEL_FROM_BOT_OAUTH_SCOPE=GovernmentConstants.TO_CHANNEL_FROM_BOT_OAUTH_SCOPE
        )
        downloader = TeamsAttachmentDownloader(
            TeamsAttachmentDownloaderOptions(
                bot_app_id="botAppId",
                adapter=TeamsAdapter(
                    config,
                    credentials_factory=MockCredentialsFactory(),
                ),
            )
        )

        input_files = await downloader.download_files(context)

        assert mock_get.call_count == 1
        self.assertEqual(
            input_files,
            [
                InputFile(
                    content=mocked_content,
                    content_type="image/png",
                    content_url="https://example.com/file.png",
                )
            ],
        )

    @mock.patch("aiohttp.ClientSession.get")
    async def test_should_download_file_with_enabled_auth_gov(self, mock_get):
        response_obj = MockedResponse()
        mock_get.return_value.__aenter__.return_value = response_obj
        mocked_content = await response_obj.read()
        attachment = Attachment(
            content_url="https://example.com/file.png", content_type="image/png", name="file.png"
        )

        context = self.create_mock_context()
        context.activity.type = "message"
        context.activity.text = "Here is the attachment"
        context.activity.attachments = [attachment]
        downloader = TeamsAttachmentDownloader(
            TeamsAttachmentDownloaderOptions(
                bot_app_id="botAppId",
                adapter=TeamsAdapter(
                    MockConfiguration(
                        GovernmentConstants.TO_CHANNEL_FROM_BOT_LOGIN_URL,
                        GovernmentConstants.TO_CHANNEL_FROM_BOT_OAUTH_SCOPE,
                        CallerIdConstants.us_gov_channel,
                    ),
                    credentials_factory=MockCredentialsFactory(),
                ),
            )
        )

        input_files = await downloader.download_files(context)

        assert mock_get.call_count == 1
        self.assertEqual(
            input_files,
            [
                InputFile(
                    content=mocked_content,
                    content_type="image/png",
                    content_url="https://example.com/file.png",
                )
            ],
        )

    @mock.patch("aiohttp.ClientSession.get")
    async def test_should_download_file_with_enabled_auth_no_audience(self, mock_get):
        response_obj = MockedResponse()
        mock_get.return_value.__aenter__.return_value = response_obj
        mocked_content = await response_obj.read()
        attachment = Attachment(
            content_url="https://example.com/file.png", content_type="image/png", name="file.png"
        )

        context = self.create_mock_context()
        context.activity.type = "message"
        context.activity.text = "Here is the attachment"
        context.activity.attachments = [attachment]
        config = MockConfiguration(
            TO_CHANNEL_FROM_BOT_LOGIN_URL=GovernmentConstants.TO_CHANNEL_FROM_BOT_LOGIN_URL,
            CALLER_ID=CallerIdConstants.us_gov_channel,
        )
        downloader = TeamsAttachmentDownloader(
            TeamsAttachmentDownloaderOptions(
                bot_app_id="botAppId",
                adapter=TeamsAdapter(
                    config,
                    credentials_factory=MockCredentialsFactory(),
                ),
            )
        )

        input_files = await downloader.download_files(context)

        assert mock_get.call_count == 1
        self.assertEqual(
            input_files,
            [
                InputFile(
                    content=mocked_content,
                    content_type="image/png",
                    content_url="https://example.com/file.png",
                )
            ],
        )



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Awaitable, Callable, Generic, Optional, TypeVar, cast

from botbuilder.core import CloudAdapterBase, TurnContext
from botbuilder.schema import Activity, TokenResponse
from botframework.connector.auth import UserTokenClient

from ..state import TurnState
from .sign_in_response import SignInResponse

StateT = TypeVar("StateT", bound=TurnState)


class Auth(ABC, Generic[StateT]):
    """
    handles user sign-in and sign-out
    """

    _on_sign_in_success: Optional[Callable[[TurnContext, StateT], Awaitable[None]]] = None
    _on_sign_in_failure: Optional[
        Callable[[TurnContext, StateT, SignInResponse], Awaitable[None]]
    ] = None

    @abstractmethod
    def is_sign_in_activity(self, activity: Activity) -> bool:
        """
        Whether the current activity is a valid activity that supports authentication

        Args:
            context (TurnContext): Context for the current turn of conversation with the user.

        Returns:
            True if valid, otherwise False.
        """

    @abstractmethod
    async def sign_in(self, context: TurnContext, state: StateT) -> Optional[str]:
        """
        Signs in a user.
        This method will be called automatically by the Application class.

        Args:
            context (TurnContext): Context for the current turn of conversation with the user.
            state (StateT): Application state

        Returns:
            The authentication token if user is signed in. Otherwise returns null.
            In that case the bot will attempt to sign the user in.
        """

    @abstractmethod
    async def sign_out(self, context: TurnContext, state: StateT) -> None:
        """
        Signs out a user.

        Args:
            context (TurnContext): Context for the current turn of conversation with the user.
            state (StateT): Application state
        """

    @abstractmethod
    async def get_token(self, context: TurnContext) -> Optional[str]:
        """
        Check if the user is signed, if they are then return the token.

        Args:
            context (TurnContext): Context for the current turn of conversation with the user.

        Returns:
            The token if the user is signed. Otherwise null.
        """

    @abstractmethod
    async def verify_state(self, context: TurnContext, state: StateT) -> Optional[TokenResponse]:
        """
        Called on signin/verifyState activity.

        Args:
            context (TurnContext): Context for the current turn of conversation with the user.
            state (StateT): Current turn state.
        """

    @abstractmethod
    async def exchange_token(self, context: TurnContext, state: StateT) -> Optional[TokenResponse]:
        """
        Called on signin/tokenExchange activity.

        Args:
            context (TurnContext): Context for the current turn of conversation with the user.
            state (StateT): Current turn state.
        """

    def on_sign_in_success(
        self, callback: Callable[[TurnContext, StateT], Awaitable[None]]
    ) -> None:
        """
        The callback is called when the user has successfully signed in.

        Args:
            callback (Callable): The callback to call.
        """
        self._on_sign_in_success = callback

    def on_sign_in_failure(
        self, callback: Callable[[TurnContext, StateT, SignInResponse], Awaitable[None]]
    ) -> None:
        """
        The callback is called when the user has failed to signed in.

        Args:
            callback (Callable): The callback to call.
        """
        self._on_sign_in_failure = callback

    def _user_token_client(self, context: TurnContext) -> UserTokenClient:
        return cast(
            UserTokenClient,
            context.turn_state.get(cast(CloudAdapterBase, context.adapter).USER_TOKEN_CLIENT_KEY),
        )



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import List, Literal, Optional


# pylint: disable=too-many-instance-attributes
@dataclass
class CompletionConfig:
    """
    Interface for the completion configuration portion of a prompt template.

    Attributes:
        completion_type (Optional[Literal['chat','text']]): Type of completion to use.
          Defaults to using the completion type of the configured default model.
          New in schema version 1.1.

        frequency_penalty (float): The models frequency_penalty as a number between 0 and 1.
          Defaults to 0.

        include_history (bool): If true, the prompt will be augmented with the conversation history.
          Defaults to True.
          New in schema version 1.1.

        include_input (bool): If true, the prompt will be augmented with the user's input.
          Defaults to True.
          New in schema version 1.1.

        include_images (bool): If true, the prompt will be augmented with any images
          uploaded by the user. Defaults to False.
          New in schema version 1.1.

        max_tokens (int): The maximum number of tokens to generate.
          Defaults to 150.

        max_input_tokens (int): The maximum number of tokens allowed in the input.
          Defaults to 2048.
          New in schema version 1.1.

        model (Optional[str]): Name of the model to use otherwise the configured
          default model is used. Defaults to None.
          New in schema version 1.1.

        presence_penalty (float): The model's presence_penalty as a number between 0 and 1.
          Defaults to 0.

        stop_sequences (Optional[List[str]]): Array of stop sequences that when hit will
          stop generation. Defaults to None.

        temperature (float): The model's temperature as a number between 0 and 2.
          Defaults to 0.

        top_p (float): The model's top_p as a number between 0 and 2.
          Defaults to 0.
    """

    completion_type: Optional[Literal["chat", "text"]] = None
    frequency_penalty: float = 0
    include_history: bool = True
    include_input: bool = True
    include_images: bool = False
    max_tokens: int = 150
    max_input_tokens: int = 2048
    model: Optional[str] = None
    presence_penalty: float = 0
    stop_sequences: Optional[List[str]] = None
    temperature: float = 0
    top_p: float = 0

    @classmethod
    def from_dict(cls, data: dict) -> "CompletionConfig":
        return cls(
            completion_type=data.get("completion_type"),
            frequency_penalty=data.get("frequency_penalty", 0),
            include_history=data.get("include_history", True),
            include_input=data.get("include_input", True),
            include_images=data.get("include_images", False),
            max_tokens=data.get("max_tokens", 150),
            max_input_tokens=data.get("max_input_tokens", 2048),
            model=data.get("model"),
            presence_penalty=data.get("presence_penalty", 0),
            stop_sequences=data.get("stop_sequences"),
            temperature=data.get("temperature", 0),
            top_p=data.get("top_p", 0),
        )



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from enum import Enum


class ActionTypes(str, Enum):
    STOP = "STOP"
    UNKNOWN_ACTION = "___UnknownAction___"
    FLAGGED_INPUT = "___FlaggedInput___"
    FLAGGED_OUTPUT = "___FlaggedOutput___"
    HTTP_ERROR = "___HttpError___"
    PLAN_READY = "___PlanReady___"
    DO_COMMAND = "___DO___"
    SAY_COMMAND = "___SAY___"
    TOO_MANY_STEPS = "__TooManySteps__"



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from botbuilder.core import TurnContext

from ...state import MemoryBase
from ..models.prompt_response import PromptResponse
from ..tokenizers import Tokenizer
from .prompt_response_validator import PromptResponseValidator
from .validation import Validation


class DefaultResponseValidator(PromptResponseValidator):
    """
    Default response validator that always returns true.
    """

    async def validate_response(
        self,
        context: TurnContext,
        memory: MemoryBase,
        tokenizer: Tokenizer,
        response: PromptResponse,
        remaining_attempts: int,
    ) -> Validation:
        return Validation()



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from unittest import IsolatedAsyncioTestCase, mock

import pytest
from botbuilder.core import TurnContext
from botbuilder.schema import Activity, ChannelAccount, ConversationAccount

from teams import Application
from tests.utils import SimpleAdapter


class TestAdaptiveCards(IsolatedAsyncioTestCase):
    application: Application

    @pytest.fixture(autouse=True)
    def before_each(self):
        self.application = Application()
        yield

    @pytest.mark.asyncio
    async def test_fetch(self):
        handler = mock.AsyncMock()  # mock handler
        self.application.task_modules.fetch("CustomForm")(handler)
        self.assertEqual(len(self.application._routes), 1)

        # The handler will be tirggered
        await self.application.on_turn(
            TurnContext(
                SimpleAdapter(),
                self.mock_activity(
                    name="task/fetch",
                    value={
                        "data": {"type": "task/fetch", "verb": "CustomForm"},
                        "context": {"theme": "default"},
                    },
                ),
            )
        )

        # The handler will not be tirggered
        await self.application.on_turn(
            TurnContext(
                SimpleAdapter(),
                self.mock_activity(
                    name="task/fetch",
                    value={
                        "data": {"type": "task/fetch", "verb": "AnotherForm"},
                        "context": {"theme": "default"},
                    },
                ),
            )
        )

        handler.assert_called_once()

    def mock_activity(self, name, value):
        return Activity(
            id="1234",
            type="invoke",
            name=name,
            value=value,
            from_property=ChannelAccount(id="user", name="Task Modules Test User"),
            recipient=ChannelAccount(id="bot", name="Task Modules Test Bot"),
            conversation=ConversationAccount(id="convo", name="Task Modules Test Convo"),
            channel_id="UnitTest",
            locale="en-uS",
            service_url="https://example.org",
        )



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from .format_citations_response import format_citations_response
from .get_used_citations import get_used_citations

__all__ = ["format_citations_response", "get_used_citations"]



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import List, Literal, Optional, Union

EmbeddingsResponseStatus = Literal[
    "success",  # The embeddings were successfully created.
    "error",  # An error occurred while creating the embeddings.
    "rate_limited",  # The request was rate limited.
]


@dataclass
class EmbeddingsResponse:
    """
    Response returned for embeddings.
    """

    status: EmbeddingsResponseStatus
    "Status of the embeddings response."

    output: Optional[Union[List[List[float]], str]] = None
    "Optional. Embeddings for the given inputs or the error string."

    message: Optional[str] = None
    "Optional. Message status."



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from unittest import TestCase

from teams.ai.prompts import SystemMessage


class TestSystemMessage(TestCase):
    def test_init(self):
        system_message = SystemMessage("template")
        self.assertEqual(system_message.template, "template")
        self.assertEqual(system_message.tokens, -1)
        self.assertEqual(system_message.text_prefix, "")
        self.assertEqual(system_message.role, "system")
        self.assertTrue(system_message.required)
        self.assertEqual(system_message.separator, "\n")

    def test_init_with_optional_params(self):
        system_message = SystemMessage("template", 100)
        self.assertEqual(system_message.template, "template")
        self.assertEqual(system_message.tokens, 100)
        self.assertEqual(system_message.text_prefix, "")
        self.assertEqual(system_message.role, "system")
        self.assertTrue(system_message.required)
        self.assertEqual(system_message.separator, "\n")



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from typing import List

from tiktoken import Encoding, get_encoding

from .tokenizer import Tokenizer


class GPTTokenizer(Tokenizer):
    """Used to encode and decode text for GPT-3.5/GPT-4 model."""

    _encoding: Encoding

    def __init__(self):
        """Initializes the GPTTokenizer object."""
        self._encoding = get_encoding("cl100k_base")

    def decode(self, tokens: List[int]) -> str:
        """Decodes a list of tokens into a string.

        Args:
            tokens (List[int]): The list of tokens to be decoded.

        Returns:
            str: The decoded string.
        """
        return self._encoding.decode(tokens)

    def encode(self, text: str) -> List[int]:
        """Encodes a string into a list of tokens.

        Args:
            text (str): The string to be encoded.

        Returns:
            List[int]: The list of encoded tokens.
        """
        return self._encoding.encode(text)



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from .conversation_state import ConversationState
from .memory import Memory, MemoryBase
from .state import State, StatePropertyAccessor, state
from .temp_state import TempState
from .todict import todict
from .turn_state import TurnState
from .user_state import UserState

__all__ = [
    "ConversationState",
    "Memory",
    "MemoryBase",
    "state",
    "State",
    "StatePropertyAccessor",
    "TurnState",
    "UserState",
    "TempState",
    "todict",
]



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from typing import List, Literal, Union
from unittest import IsolatedAsyncioTestCase, mock

import httpx
import openai

from teams.ai.embeddings import OpenAIEmbeddings, OpenAIEmbeddingsOptions

embedding_1 = [
    -0.006929283495992422,
    -0.016709936782717705,
    0.009889421984553337,
    -0.0007253705407492816,
]
embedding_2 = [
    -0.006929283495992422,
    -0.016709936782717705,
    0.009889421984553337,
    -0.0007253705407492816,
]


class MockAsyncEmbeddings:
    async def create(
        self,
        input: Union[str, List[str], List[int], List[List[int]]],
        model: Union[str, Literal["text-embedding-ada-002"]],
    ) -> openai.types.CreateEmbeddingResponse:
        # pylint: disable=unused-argument
        # this is necessary to override and mock the class
        return openai.types.CreateEmbeddingResponse(
            data=[
                openai.types.Embedding(embedding=embedding_1, index=1, object="embedding"),
                openai.types.Embedding(embedding=embedding_2, index=0, object="embedding"),
            ],
            model="text-embedding-ada-002",
            object="list",
            usage=openai.types.create_embedding_response.Usage(prompt_tokens=5, total_tokens=5),
        )


class MockAsyncOpenAI:
    embeddings = MockAsyncEmbeddings()


class MockAsyncEmbeddingsRateLimited:
    async def create(
        self,
        input: Union[str, List[str], List[int], List[List[int]]],
        model: Union[str, Literal["text-embedding-ada-002"]],
    ) -> openai.types.CreateEmbeddingResponse:
        raise openai.RateLimitError(
            message="This is a rate limited error",
            response=httpx.Response(
                status_code=429, request=httpx.Request(method="method", url="url")
            ),
            body=None,
        )


class MockAsyncOpenAIRateLimited:
    embeddings = MockAsyncEmbeddingsRateLimited()


class MockAsyncEmbeddingsAPIError:
    async def create(
        self,
        input: Union[str, List[str], List[int], List[List[int]]],
        model: Union[str, Literal["text-embedding-ada-002"]],
    ) -> openai.types.CreateEmbeddingResponse:
        raise openai.APIError(
            message="This is a bad request error",
            request=httpx.Request(method="method", url="url"),
            body=None,
        )


class MockAsyncOpenAIAPIError:
    embeddings = MockAsyncEmbeddingsAPIError()


class TestOpenAIEmbeddings(IsolatedAsyncioTestCase):
    options: OpenAIEmbeddingsOptions
    embeddings: OpenAIEmbeddings
    options_with_array: OpenAIEmbeddingsOptions
    embeddings_with_array: OpenAIEmbeddings

    def setUp(self):
        self.options = OpenAIEmbeddingsOptions(
            api_key="empty", model="text-embedding-ada-002", log_requests=True, retry_policy=[2, 4]
        )
        self.options_with_array = OpenAIEmbeddingsOptions(
            api_key="empty",
            model="text-embedding-ada-002",
            organization="random",
            log_requests=False,
            request_config={
                "Authorization": "Bearer empty",
                "Content-Type": "application/json",
                "User-Agent": "123",
            },
        )

    @mock.patch("openai.AsyncOpenAI", return_value=MockAsyncOpenAI)
    async def test_string_embedding(self, mock_async_open_ai):
        self.embeddings = OpenAIEmbeddings(self.options)
        section = await self.embeddings.create_embeddings("This is an embedding")
        self.assertTrue(mock_async_open_ai.called)
        self.assertEqual(section.status, "success")
        self.assertEqual(section.output, [embedding_2, embedding_1])

    @mock.patch("openai.AsyncOpenAI", return_value=MockAsyncOpenAI)
    async def test_array_embedding(self, mock_async_open_ai):
        self.embeddings_with_array = OpenAIEmbeddings(self.options_with_array)
        section = await self.embeddings_with_array.create_embeddings(["This is", "an embedding"])
        self.assertTrue(mock_async_open_ai.called)
        self.assertEqual(section.status, "success")
        self.assertEqual(section.output, [embedding_2, embedding_1])

    @mock.patch("openai.AsyncOpenAI", return_value=MockAsyncOpenAIAPIError)
    async def test_api_error(self, mock_async_open_ai):
        self.embeddings = OpenAIEmbeddings(self.options)
        section = await self.embeddings.create_embeddings("This is an embedding")
        self.assertTrue(mock_async_open_ai.called)
        self.assertEqual(section.status, "error")
        self.assertEqual(
            section.output,
            "The embeddings API returned an error status of None: This is a bad request error",
        )

    @mock.patch("openai.AsyncOpenAI", return_value=MockAsyncOpenAIRateLimited)
    async def test_rate_limited(self, mock_async_open_ai):
        self.embeddings = OpenAIEmbeddings(self.options)
        section = await self.embeddings.create_embeddings("This is an embedding")
        self.assertTrue(mock_async_open_ai.called)
        self.assertEqual(section.status, "rate_limited")
        self.assertEqual(section.output, "The embeddings API returned a rate limit error.")



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Awaitable, Callable, List

from botbuilder.core import TurnContext

from ...state import MemoryBase
from ..tokenizers import Tokenizer

PromptFunction = Callable[
    [TurnContext, MemoryBase, "PromptFunctions", Tokenizer, List[str]], Awaitable[Any]
]
"""
A function that can be called from a prompt template string.

Parameters:
    context (TurnContext): Context for the current turn of conversation.
    memory (MemoryBase): Interface used to access state variables.
    functions (PromptFunctions): Collection of functions that can be called
      from a prompt template string.
    tokenizer (Tokenizer): Tokenizer used to encode/decode strings.
    args (List[str]): Arguments to the function as an array of strings.

Returns:
    Any: The return type is not specified, so it can be any type.
"""


class PromptFunctions(ABC):
    """A collection of functions that can be called from a prompt template string."""

    @abstractmethod
    def has_function(self, name: str) -> bool:
        """
        Checks if a function is defined.

        Args:
            name (str): Name of the function to lookup.

        Returns:
            bool: True if the function is defined, False otherwise.
        """

    @abstractmethod
    def get_function(self, name: str) -> PromptFunction:
        """
        Looks up a given function.

        Args:
            name (str): Name of the function to lookup.

        Raises:
            Error: If the function is not defined.

        Returns:
            PromptFunction: The function associated with the given name.
        """

    @abstractmethod
    async def invoke_function(
        self,
        name: str,
        context: TurnContext,
        memory: MemoryBase,
        tokenizer: Tokenizer,
        args: List[str],
    ):
        """
        Calls the given function.

        Args:
            name (str): Name of the function to invoke.
            context (TurnContext): Context for the current turn of conversation.
            memory (MemoryBase): Interface used to access state variables.
            tokenizer (Tokenizer): Tokenizer used to encode/decode strings.
            args (List[str]): Arguments to pass to the function as an array of strings.

        Raises:
            Error: If the function is not defined.
        """



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

import json
from dataclasses import asdict
from typing import List

from botbuilder.core import TurnContext

from ...state import MemoryBase
from ..tokenizers import Tokenizer
from .function_call import FunctionCall
from .message import Message
from .prompt_functions import PromptFunctions
from .rendered_prompt_section import RenderedPromptSection
from .sections.prompt_section_base import PromptSectionBase


class FunctionCallMessage(PromptSectionBase):
    """
    An `assistant` message containing a function to call.

    The function call information is returned by the model so we use an "assistant" message to
    represent it in conversation history.
    """

    _length: int
    _function_call: FunctionCall

    def __init__(
        self, function_call: FunctionCall, tokens: float = -1, assistant_prefix: str = "assistant: "
    ):
        """
        Creates a new `FunctionCallMessage` instance.

        Args:
            function_call (FunctionCall): Name and arguments of the function to call.
            tokens (int, optional): Sizing strategy for this section. Defaults to `auto`.
            assistant_prefix (str, optional): Prefix to use for assistant messages when
              rendering as text. Defaults to `assistant: `.
        """
        super().__init__(tokens, True, "\n", assistant_prefix)
        self._length: int = -1
        self._function_call: FunctionCall = function_call

    @property
    def function_call(self) -> FunctionCall:
        """
        Name and arguments of the function to call.
        """
        return self._function_call

    async def render_as_messages(
        self,
        context: TurnContext,
        memory: "MemoryBase",
        functions: PromptFunctions,
        tokenizer: Tokenizer,
        max_tokens: int,
    ) -> RenderedPromptSection[List[Message]]:
        """
        Renders the section as a list of messages.

        Args:
            context (TurnContext): Context for the current turn of conversation with the user.
            memory (MemoryBase): An interface for accessing state values.
            functions (PromptFunctions): Registry of functions that can be used by the section.
            tokenizer (Tokenizer): Tokenizer to use when rendering the section.
            max_tokens (int): Maximum number of tokens allowed to be rendered.

        Returns:
            RenderedPromptSection[List[Message]]: The rendered prompt section as a list of messages.
        """
        # Calculate and cache response text and length
        if self._length < 0:
            self._length = len(tokenizer.encode(json.dumps(asdict(self.function_call))))

        # Return output
        return self._return_messages(
            [Message("assistant", function_call=self.function_call)],
            self._length,
            tokenizer,
            max_tokens,
        )



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from unittest import IsolatedAsyncioTestCase
from unittest.mock import MagicMock

from botbuilder.core import TurnContext

from teams.ai.prompts import ConversationHistorySection, Message, PromptFunctions
from teams.ai.tokenizers import GPTTokenizer
from teams.state import ConversationState, TempState, TurnState, UserState


class TestConversationHistory(IsolatedAsyncioTestCase):
    async def asyncSetUp(self) -> None:
        self.prompt_functions = MagicMock(spec=PromptFunctions)
        self.turn_context = MagicMock(spec=TurnContext)
        self.memory = await TurnState[ConversationState, UserState, TempState].load(
            self.turn_context
        )
        self.memory.conversation["history"] = [
            Message("user", "Hello"),
            Message("assistant", "Hi! How can I help you?"),
            Message("user", "I'd like to book a flight"),
            Message("assistant", "Sure, where would you like to go?"),
        ]

    def test_init(self):
        conversation_history = ConversationHistorySection("conversation.history")
        self.assertEqual(conversation_history.tokens, 1.0)
        self.assertFalse(conversation_history.required)
        self.assertEqual(conversation_history.separator, "\n")
        self.assertEqual(conversation_history.user_prefix, "user: ")
        self.assertEqual(conversation_history.assistant_prefix, "assistant: ")

    def test_init_with_optional_params(self):
        conversation_history = ConversationHistorySection(
            "conversation.history",
            tokens=0.5,
            required=True,
            user_prefix="user: ",
            assistant_prefix="assistant: ",
            separator="\n",
        )
        self.assertEqual(conversation_history.tokens, 0.5)
        self.assertTrue(conversation_history.required)
        self.assertEqual(conversation_history.separator, "\n")
        self.assertEqual(conversation_history.user_prefix, "user: ")
        self.assertEqual(conversation_history.assistant_prefix, "assistant: ")

    async def test_render_as_text(self):
        conversation_history = ConversationHistorySection("conversation.history")
        result = await conversation_history.render_as_text(
            self.turn_context, self.memory, self.prompt_functions, GPTTokenizer(), 100
        )
        self.assertEqual(
            result.output,
            'user: "Hello"\n'
            'assistant: "Hi! How can I help you?"\n'
            'user: "I\'d like to book a flight"\n'
            'assistant: "Sure, where would you like to go?"',
        )
        self.assertEqual(result.length, 42)
        self.assertFalse(result.too_long)

    async def test_render_as_text_include_initial_line_when_required(self):
        conversation_history = ConversationHistorySection("conversation.history", required=True)
        result = await conversation_history.render_as_text(
            self.turn_context, self.memory, self.prompt_functions, GPTTokenizer(), 1
        )
        self.assertEqual(result.output, 'assistant: "Sure, where would you like to go?"')
        self.assertEqual(result.length, 12)
        self.assertTrue(result.too_long)

    async def test_render_as_text_truncate_history(self):
        conversation_history = ConversationHistorySection("conversation.history")
        result = await conversation_history.render_as_text(
            self.turn_context, self.memory, self.prompt_functions, GPTTokenizer(), 12
        )
        self.assertEqual(result.output, 'assistant: "Sure, where would you like to go?"')
        self.assertEqual(result.length, 12)
        self.assertFalse(result.too_long)

    async def test_render_as_text_empty_history(self):
        conversation_history = ConversationHistorySection("conversation.no_history")
        result = await conversation_history.render_as_text(
            self.turn_context, self.memory, self.prompt_functions, GPTTokenizer(), 100
        )
        self.assertEqual(result.output, "")
        self.assertEqual(result.length, 0)
        self.assertFalse(result.too_long)

    async def test_render_as_text_long_last_message(self):
        conversation_history = ConversationHistorySection("conversation.history")
        result = await conversation_history.render_as_text(
            self.turn_context, self.memory, self.prompt_functions, GPTTokenizer(), 1
        )
        self.assertEqual(result.output, "")
        self.assertEqual(result.length, 0)
        self.assertFalse(result.too_long)

    async def test_render_as_messages(self):
        conversation_history = ConversationHistorySection("conversation.history")
        result = await conversation_history.render_as_messages(
            self.turn_context, self.memory, self.prompt_functions, GPTTokenizer(), 100
        )
        self.assertEqual(
            result.output,
            [
                Message("user", '"Hello"'),
                Message("assistant", '"Hi! How can I help you?"'),
                Message("user", '"I\'d like to book a flight"'),
                Message("assistant", '"Sure, where would you like to go?"'),
            ],
        )
        self.assertEqual(result.length, 30)
        self.assertFalse(result.too_long)

    async def test_render_as_messages_include_initial_line_when_required(self):
        conversation_history = ConversationHistorySection("conversation.history", required=True)
        result = await conversation_history.render_as_messages(
            self.turn_context, self.memory, self.prompt_functions, GPTTokenizer(), 1
        )
        self.assertEqual(
            result.output, [Message("assistant", '"Sure, where would you like to go?"')]
        )
        self.assertEqual(result.length, 10)
        self.assertTrue(result.too_long)

    async def test_render_as_messages_truncate_history(self):
        conversation_history = ConversationHistorySection("conversation.history")
        result = await conversation_history.render_as_messages(
            self.turn_context, self.memory, self.prompt_functions, GPTTokenizer(), 12
        )
        self.assertEqual(
            result.output, [Message("assistant", '"Sure, where would you like to go?"')]
        )
        self.assertEqual(result.length, 10)
        self.assertFalse(result.too_long)

    async def test_render_as_messages_empty_history(self):
        conversation_history = ConversationHistorySection("conversation.no_history")
        result = await conversation_history.render_as_messages(
            self.turn_context, self.memory, self.prompt_functions, GPTTokenizer(), 100
        )
        self.assertEqual(result.output, [])
        self.assertEqual(result.length, 0)
        self.assertFalse(result.too_long)

    async def test_render_as_messages_long_last_message(self):
        conversation_history = ConversationHistorySection("conversation.history")
        result = await conversation_history.render_as_messages(
            self.turn_context, self.memory, self.prompt_functions, GPTTokenizer(), 1
        )
        self.assertEqual(result.output, [])
        self.assertEqual(result.length, 0)
        self.assertFalse(result.too_long)



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from __future__ import annotations

from threading import Timer
from typing import Optional

from botbuilder.core import TurnContext
from botbuilder.schema import Activity, ActivityTypes, ErrorResponseException


class Typing:
    """
    Encapsulates the logic for sending "typing" activity to the user.
    """

    _interval: int
    _timer: Optional[Timer] = None

    def __init__(self, interval=1000) -> None:
        self._interval = interval

    async def start(self, context: TurnContext) -> None:
        if self._timer is not None:
            return

        func = self._on_timer(context)
        self._timer = Timer(self._interval, func)
        self._timer.start()
        await func()

    def stop(self) -> None:
        if self._timer:
            self._timer.cancel()
            self._timer = None

    def _on_timer(self, context: TurnContext):
        async def __call__():
            try:
                await context.send_activity(Activity(type=ActivityTypes.typing))
            except ErrorResponseException:
                self.stop()

        return __call__
